"""ggml-python implemention of the CLIP model
"""
import io
import os
import ctypes
import struct
import argparse
import numpy as np
from typing import List, Tuple, Dict
import ggml
from ggml.experimental import GGML_FTYPE, Context, InitParams, Tensor, GGML_TYPE, CGraph


def compute_ctx_size(fin: io.BufferedReader) -> int:
    # Save current position in file and get file size, then return
    position = fin.tell()

    ctx_size = 0
    while True:
        nbytes = struct.calcsize("iii")
        data = fin.read(nbytes)
        if len(data) != nbytes:
            break
        (n_dims, s_len, ftype) = struct.unpack("iii", data)
        dims = struct.unpack("i" * n_dims, fin.read(struct.calcsize("i" * n_dims)))
        if ftype == 0:
            _format = "f"
        if ftype == 1:
            _format = "e"
        n_bytes = struct.calcsize(_format * int(np.prod(dims)))
        ctx_size += n_bytes
        ctx_size += 256  # Padding?
        name = fin.read(s_len).decode("utf-8")
        # print(f"Name: {name}, dims: {dims}, n_bytes: {n_bytes}")

        fin.seek(n_bytes, os.SEEK_CUR)

    # Seek back to saved position
    fin.seek(position)
    return ctx_size


class ResidualAttentionBlock:
    def __init__(
        self,
        ctx: Context,
        wtype: GGML_TYPE,
        embed_dim: int,
        heads: int,
        use_attn_mask: bool = False,
    ):
        self.tensors: Dict[str, Tensor] = {}
        self.n_head = heads
        self.embed_dim = embed_dim
        self.use_attn_mask = use_attn_mask
        # Layer Norm 1 (ln_1)
        self.ln_1_weight = Tensor.new_tensor_1d(wtype, embed_dim, ctx=ctx)
        self.ln_1_bias = Tensor.new_tensor_1d(wtype, embed_dim, ctx=ctx)
        self.tensors["ln_1.weight"] = self.ln_1_weight
        self.tensors["ln_1.bias"] = self.ln_1_bias

        # Attention Block (attn)
        self.in_proj_weight = Tensor.new_tensor_2d(
            wtype, embed_dim, 3 * embed_dim, ctx=ctx
        )
        self.in_proj_bias = Tensor.new_tensor_1d(wtype, 3 * embed_dim, ctx=ctx)
        self.out_proj_weight = Tensor.new_tensor_2d(
            wtype, embed_dim, embed_dim, ctx=ctx
        )
        self.out_proj_bias = Tensor.new_tensor_1d(wtype, embed_dim, ctx=ctx)
        self.tensors["attn.in_proj_weight"] = self.in_proj_weight
        self.tensors["attn.in_proj_bias"] = self.in_proj_bias
        self.tensors["attn.out_proj.weight"] = self.out_proj_weight
        self.tensors["attn.out_proj.bias"] = self.out_proj_bias

        # Layer Norm 2 (ln_2)
        self.ln_2_weight = Tensor.new_tensor_1d(wtype, embed_dim, ctx=ctx)
        self.ln_2_bias = Tensor.new_tensor_1d(wtype, embed_dim, ctx=ctx)
        self.tensors["ln_2.weight"] = self.ln_2_weight
        self.tensors["ln_2.bias"] = self.ln_2_bias

        # MLP (mlp)
        self.mlp_c_fc_weight = Tensor.new_tensor_2d(
            wtype, embed_dim, embed_dim * 4, ctx=ctx
        )
        self.mlp_c_fc_bias = Tensor.new_tensor_1d(wtype, embed_dim * 4, ctx=ctx)
        self.mlp_c_proj_weight = Tensor.new_tensor_2d(
            wtype, embed_dim * 4, embed_dim, ctx=ctx
        )
        self.mlp_c_proj_bias = Tensor.new_tensor_1d(wtype, embed_dim, ctx=ctx)
        self.tensors["mlp.c_fc.weight"] = self.mlp_c_fc_weight
        self.tensors["mlp.c_fc.bias"] = self.mlp_c_fc_bias
        self.tensors["mlp.c_proj.weight"] = self.mlp_c_proj_weight
        self.tensors["mlp.c_proj.bias"] = self.mlp_c_proj_bias

    @staticmethod
    def compute_forward_mem_size(
        N: int, width: int, n_heads: int, use_attn_mask: bool = False
    ) -> int:
        e_size = 4
        ggml_overhead = 256
        mem_size = 0
        mem_size += (
            e_size * width * N + ggml_overhead
        ) * 5  # ln_1: repeat, repeat, mul, add, norm
        mem_size += (
            e_size * width * 3 * N + ggml_overhead
        ) * 3  # in_proj: mul_mat, repeat, add
        mem_size += ggml_overhead * 3  # view_2d: Qcur, Kcur, Vcur
        mem_size += (
            (e_size * (width // n_heads) * n_heads * N + ggml_overhead)
            + 2 * ggml_overhead
        ) * 2  # K,Q: new_tensor, cpy, permute
        mem_size += e_size * N * N * n_heads + ggml_overhead  # KQ
        mem_size += e_size * 4 + 256  # KQ_scaled: new_f32
        mem_size += e_size * N * N * n_heads + ggml_overhead  # KQ_scaled
        if use_attn_mask:
            mem_size += (
                e_size * N * N * n_heads + ggml_overhead + e_size * 4 + ggml_overhead
            )  # diag_mask_inf
        mem_size += e_size * N * N * n_heads + ggml_overhead  # KQ_soft_max
        mem_size += (
            e_size * (width // n_heads) * n_heads * N + ggml_overhead
        )  # V_trans: new_tensor_3d
        mem_size += ggml_overhead * 2  # V_trans: cpy and permute
        mem_size += (
            e_size * (width // n_heads) * n_heads * N + ggml_overhead
        )  # V_trans: new_tensor_3d
        mem_size += ggml_overhead  # V_trans: cpy
        mem_size += (
            e_size * (width // n_heads) * n_heads * N + ggml_overhead
        )  # KQV: mul_mat
        mem_size += ggml_overhead  # KQV_merged: permute
        mem_size += e_size * width * N + ggml_overhead  # KQV_merged: new_tensor_2d
        mem_size += ggml_overhead  # KQV_merged: cpy
        mem_size += (
            e_size * width * N + ggml_overhead
        ) * 3  # out_proj: mul_mat, repeat, add
        mem_size += e_size * width * N + ggml_overhead  # Add residual
        mem_size += (
            e_size * width * N + ggml_overhead
        ) * 5  # ln_2: norm, add, repeat, repeat, mul
        mem_size += (
            e_size * width * 4 * N + ggml_overhead
        ) * 3  # MLP: mul_mat, repeat, add
        mem_size += (e_size * 4 + 256) * 2  # SiLU: sf_in, sf_out
        mem_size += (
            e_size * width * 4 * N + ggml_overhead
        ) * 3  # SiLU: scale, silu, scale
        mem_size += (
            e_size * width * N + ggml_overhead
        ) * 3  # mlp_c_proj: mul_mat, repeat, add
        mem_size += e_size * width * N + ggml_overhead  # Add Residual

        return mem_size

    def forward(self, inpL: Tensor, ctx: Context, gf: CGraph) -> Tensor:
        N = inpL.shape[1]

        # [768, N]
        cur = Tensor.norm(inpL, ctx=ctx)
        # cur = ln_1_weight * cur + ln_1_bias
        # [768, N]
        cur = Tensor.add(
            Tensor.mul(Tensor.repeat(self.ln_1_weight, cur, ctx=ctx), cur, ctx=ctx),
            Tensor.repeat(self.ln_1_bias, cur, ctx=ctx),
            ctx=ctx,
        )

        # cur = in_proj_weight * cur + in_proj_bias
        # [768, N] - cur (in)
        # [2304, 768] - in_proj_weight
        # [2304, 1] - in_proj_bias
        # [2304, N] - cur (out)
        cur = Tensor.mul_mat(self.in_proj_weight, cur, ctx=ctx)

        cur = Tensor.add(Tensor.repeat(self.in_proj_bias, cur, ctx=ctx), cur, ctx=ctx)

        # Self-Attention
        n_embd = cur.shape[0] // 3

        Qcur = Tensor.view_2d(
            cur,
            n_embd,
            N,
            cur.tensor.contents.nb[1],
            0 * ctypes.sizeof(ctypes.c_float) * n_embd,
            ctx=ctx,
        )

        Kcur = Tensor.view_2d(
            cur,
            n_embd,
            N,
            cur.tensor.contents.nb[1],
            1 * ctypes.sizeof(ctypes.c_float) * n_embd,
            ctx=ctx,
        )

        Vcur = Tensor.view_2d(
            cur,
            n_embd,
            N,
            cur.tensor.contents.nb[1],
            2 * ctypes.sizeof(ctypes.c_float) * n_embd,
            ctx=ctx,
        )

        Q = Tensor.permute(
            Tensor.cpy(
                Qcur,
                Tensor.new_tensor_3d(
                    GGML_TYPE.F32, n_embd // self.n_head, self.n_head, N, ctx=ctx
                ),
                ctx=ctx,
            ),
            0,
            2,
            1,
            3,
            ctx=ctx,
        )

        K = Tensor.permute(
            Tensor.cpy(
                Kcur,
                Tensor.new_tensor_3d(
                    GGML_TYPE.F32, n_embd // self.n_head, self.n_head, N, ctx=ctx
                ),
                ctx=ctx,
            ),
            0,
            2,
            1,
            3,
            ctx=ctx,
        )

        KQ = Tensor.mul_mat(K, Q, ctx=ctx)

        KQ_scaled = Tensor.scale(
            KQ,
            Tensor.new_f32(
                1.0 / np.sqrt(float(n_embd) / self.n_head),
                ctx=ctx,
            ),
            ctx=ctx,
        )
        if self.use_attn_mask:
            KQ_masked = Tensor.diag_mask_inf(KQ_scaled, 0, ctx=ctx)
            KQ_soft_max = Tensor.soft_max(KQ_masked, ctx=ctx)
        else:
            KQ_soft_max = Tensor.soft_max(KQ_scaled, ctx=ctx)

        V_trans = Tensor.cpy(
            Tensor.permute(
                Tensor.cpy(
                    Vcur,
                    Tensor.new_tensor_3d(
                        GGML_TYPE.F32, n_embd // self.n_head, self.n_head, N, ctx=ctx
                    ),
                    ctx=ctx,
                ),
                1,
                2,
                0,
                3,
                ctx=ctx,
            ),
            Tensor.new_tensor_3d(
                GGML_TYPE.F32, N, n_embd // self.n_head, self.n_head, ctx=ctx
            ),
            ctx=ctx,
        )

        KQV = Tensor.mul_mat(V_trans, KQ_soft_max, ctx=ctx)

        KQV_merged = Tensor.permute(
            KQV,
            0,
            2,
            1,
            3,
            ctx=ctx,
        )

        cur = Tensor.cpy(
            KQV_merged,
            Tensor.new_tensor_2d(
                GGML_TYPE.F32,
                n_embd,
                N,
                ctx=ctx,
            ),
            ctx=ctx,
        )

        cur = Tensor.mul_mat(
            self.out_proj_weight,
            cur,
            ctx=ctx,
        )

        cur = Tensor.add(Tensor.repeat(self.out_proj_bias, cur, ctx=ctx), cur, ctx=ctx)

        # Add Residual
        inpL = Tensor.add(inpL, cur, ctx=ctx)

        # LN2
        cur = Tensor.norm(inpL, ctx=ctx)
        cur = Tensor.add(
            Tensor.mul(Tensor.repeat(self.ln_2_weight, cur, ctx=ctx), cur, ctx=ctx),
            Tensor.repeat(self.ln_2_bias, cur, ctx=ctx),
            ctx=ctx,
        )

        # MLP
        # c_fc
        cur = Tensor.mul_mat(self.mlp_c_fc_weight, cur, ctx=ctx)
        cur = Tensor.add(Tensor.repeat(self.mlp_c_fc_bias, cur, ctx=ctx), cur, ctx=ctx)

        # QuickGELU -  x * sigmoid(1.702 * x)
        cur = Tensor.scale(cur, Tensor.new_f32(1.702, ctx=ctx), ctx=ctx)

        cur = Tensor.silu(cur, ctx=ctx)

        cur = Tensor.scale(cur, Tensor.new_f32(1 / 1.702, ctx=ctx), ctx=ctx)

        # c_proj
        cur = Tensor.mul_mat(self.mlp_c_proj_weight, cur, ctx=ctx)
        cur = Tensor.add(
            Tensor.repeat(self.mlp_c_proj_bias, cur, ctx=ctx), cur, ctx=ctx
        )

        # Add Residual
        cur = Tensor.add(inpL, cur, ctx=ctx)
        return cur


class VisionTransformer:
    def __init__(
        self,
        ctx: Context,
        wtype: GGML_TYPE,
        input_resolution: int,
        patch_size: int,
        width: int,
        heads: int,
        layers: int,
        output_dim: int,
    ):
        self.layers = layers
        self.tensors: Dict[str, Tensor] = {}

        # Class Embedding (visual.class_embedding)
        self.visual_class_embedding = Tensor.new_tensor_1d(wtype, width, ctx=ctx)
        self.tensors["visual.class_embedding"] = self.visual_class_embedding

        # Positional Embedding (visual.positional_embedding)
        self.visual_positional_embedding = Tensor.new_tensor_2d(
            wtype, width, (input_resolution // patch_size) ** 2 + 1, ctx=ctx
        )
        self.tensors["visual.positional_embedding"] = self.visual_positional_embedding

        # Convolutional Layer (visual.conv1.weight)
        wtype_f16 = GGML_TYPE(ggml.ggml_ftype_to_ggml_type(ctypes.c_int(1)))
        self.visual_conv1_weight = Tensor.new_tensor_4d(
            wtype_f16, patch_size, patch_size, 3, width, ctx=ctx
        )
        self.tensors["visual.conv1.weight"] = self.visual_conv1_weight

        # pre Layer Norm Weight (visual.ln_pre.weight)
        self.visual_ln_pre_weight = Tensor.new_tensor_1d(wtype, width, ctx=ctx)
        self.tensors["visual.ln_pre.weight"] = self.visual_ln_pre_weight

        # pre Layer Norm Bias (visual.ln_pre.bias)
        self.visual_ln_pre_bias = Tensor.new_tensor_1d(wtype, width, ctx=ctx)
        self.tensors["visual.ln_pre.bias"] = self.visual_ln_pre_bias
        self.resblocks = []
        for i in range(layers):
            resblock = ResidualAttentionBlock(
                ctx=ctx, wtype=wtype, embed_dim=width, heads=heads, use_attn_mask=False
            )
            self.resblocks.append(resblock)
            self.tensors.update(
                {
                    f"visual.transformer.resblocks.{i}." + k: v
                    for k, v in resblock.tensors.items()
                }
            )

        # post Layer Norm (visual.ln_post)
        self.visual_ln_post_weight = Tensor.new_tensor_1d(wtype, width, ctx=ctx)
        self.visual_ln_post_bias = Tensor.new_tensor_1d(wtype, width, ctx=ctx)
        self.tensors["visual.ln_post.weight"] = self.visual_ln_post_weight
        self.tensors["visual.ln_post.bias"] = self.visual_ln_post_bias

        # Visual Projection (visual.proj)
        self.visual_proj = Tensor.new_tensor_2d(wtype, output_dim, width, ctx=ctx)
        self.tensors["visual.proj"] = self.visual_proj


class ClipModel:
    def __init__(
        self,
        ctx: Context,
        wtype: GGML_TYPE,
        vision_width: int,
        vision_layers: int,
        vision_patch_size: int,
        image_resolution: int,
        embed_dim: int,
        context_length: int,
        vocab_size: int,
        transformer_width: int,
        transformer_heads: int,
        transformer_layers: int,
        n_threads: int,
    ):
        self.n_threads = n_threads
        self.tensors: Dict[str, Tensor] = {}

        # Vision Transformer
        self.vision_layers = vision_layers
        self.vision_patch_size = vision_patch_size
        self.vision_width = vision_width
        self.vision_heads = vision_width // 64
        self.image_resolution = image_resolution
        self.grid_size = image_resolution // vision_patch_size

        # Text Transformer
        self.context_length = context_length
        self.transformer_width = transformer_width
        self.transformer_heads = transformer_heads
        self.transformer_layers = transformer_layers

        self.embed_dim = embed_dim

        # Positional Embedding (position_embedding)
        self.positional_embedding = Tensor.new_tensor_2d(
            wtype, transformer_width, context_length, ctx=ctx
        )
        self.tensors["positional_embedding"] = self.positional_embedding

        # Text Projection (text_projection)
        self.text_projection = Tensor.new_tensor_2d(
            wtype, transformer_width, embed_dim, ctx=ctx
        )
        self.tensors["text_projection"] = self.text_projection

        # Logit Scale (logit_scale)
        self.logit_scale = Tensor.new_tensor_1d(wtype, 1, ctx=ctx)
        self.tensors["logit_scale"] = self.logit_scale

        # Visual Transformer (visual.)
        self.visual = VisionTransformer(
            ctx=ctx,
            wtype=wtype,
            input_resolution=image_resolution,
            patch_size=vision_patch_size,
            width=vision_width,
            layers=vision_layers,
            heads=self.vision_heads,
            output_dim=embed_dim,
        )
        self.tensors.update(self.visual.tensors)

        # Transformer (transformer.)
        self.transformer_res_blocks = []
        for i in range(transformer_layers):
            res_block = ResidualAttentionBlock(
                ctx=ctx,
                wtype=wtype,
                embed_dim=transformer_width,
                heads=transformer_heads,
                use_attn_mask=True,
            )
            self.transformer_res_blocks.append(res_block)
            self.tensors.update(
                {
                    f"transformer.resblocks.{i}." + k: v
                    for k, v in res_block.tensors.items()
                }
            )

        # Token Embedding (token_embedding.weight)
        self.token_embedding = Tensor.new_tensor_2d(
            wtype, transformer_width, vocab_size, ctx=ctx
        )
        self.tensors["token_embedding.weight"] = self.token_embedding

        # Final Layer Norm (ln_final.weight)
        self.ln_final_weight = Tensor.new_tensor_1d(wtype, transformer_width, ctx=ctx)
        self.tensors["ln_final.weight"] = self.ln_final_weight

        # Final Layer Norm (ln_final.bias)
        self.ln_final_bias = Tensor.new_tensor_1d(wtype, transformer_width, ctx=ctx)
        self.tensors["ln_final.bias"] = self.ln_final_bias

    def encode_image(self, image):
        tensor = self._encode_image_internal(image)
        return tensor.numpy().copy().reshape(1, -1)

    def encode_text(self, text_embds):
        encodings = []
        # TODO: batchify
        for text_embd in text_embds:
            tensor = self._encode_text_internal(text_embd)
            encodings.append(tensor.numpy().copy().reshape(1, -1))
        return np.concatenate(encodings, axis=0)

    def __call__(self, image, text):
        image_features = self.encode_image(image)
        text_features = self.encode_text(text)

        # normalized features
        image_features = image_features / np.linalg.norm(
            image_features, axis=1, keepdims=True
        )
        text_features = text_features / np.linalg.norm(
            text_features, axis=1, keepdims=True
        )

        # cosine similarity as logits
        logit_scale = np.exp(self.logit_scale.numpy().copy())
        logits_per_image = logit_scale * image_features @ text_features.T
        logits_per_text = logits_per_image.T

        # shape = [global_batch_size, global_batch_size]
        return logits_per_image, logits_per_text

    def _text_encoder_compute_forward_memsize(self):
        mem_size = 0
        e_size = 4
        ggml_overhead = 256
        mem_size += e_size * self.context_length + ggml_overhead  # input embd

        mem_size += (
            e_size * self.context_length * self.embed_dim + ggml_overhead
        )  # token embedding

        mem_size += (
            e_size * self.context_length * self.embed_dim + ggml_overhead
        )  # add positional embedding
        res_block_mem_size = ResidualAttentionBlock.compute_forward_mem_size(
            self.context_length,
            self.transformer_width,
            self.transformer_heads,
            use_attn_mask=True,
        )
        mem_size += res_block_mem_size * self.transformer_layers
        mem_size += (
            e_size * self.transformer_width * self.context_length + ggml_overhead
        ) * 5  # ln_final

        mem_size += ggml_overhead  # view

        mem_size += e_size * self.embed_dim + ggml_overhead  # Text Proj: output
        mem_size += ggml_overhead  # Text Proj: Transpose
        mem_size += (
            e_size * self.embed_dim * self.embed_dim * ggml_overhead
        )  # Text Proj: cpy
        return mem_size

    def _encode_text_internal(self, embd_inp: np.ndarray):
        wtype = GGML_TYPE(ggml.ggml_ftype_to_ggml_type(ctypes.c_int(0)))
        N = self.context_length
        mem_size = self._text_encoder_compute_forward_memsize()
        mem_buffer = np.empty(mem_size, dtype=np.uint8)
        init_params = InitParams(
            mem_size=mem_size, mem_buffer=mem_buffer.ctypes.data_as(ctypes.c_void_p)
        )
        ctx0 = Context(init_params=init_params)

        gf = CGraph(cgraph=ggml.ggml_cgraph(n_threads=self.n_threads), ctx=ctx0)

        embd = Tensor.new_tensor_1d(GGML_TYPE.I32, N, ctx=ctx0)
        embd.numpy()[:] = np.array(embd_inp, dtype=np.int32)
        inpL = Tensor.get_rows(self.token_embedding, embd, ctx=ctx0)
        cur = Tensor.add(inpL, self.positional_embedding, ctx=ctx0)

        for il in range(self.transformer_layers):
            resblock = self.transformer_res_blocks[il]
            cur = resblock.forward(cur, ctx=ctx0, gf=gf)

        cur = Tensor.norm(cur, ctx=ctx0)
        cur = Tensor.add(
            Tensor.mul(
                Tensor.repeat(self.ln_final_weight, cur, ctx=ctx0),
                cur,
                ctx=ctx0,
            ),
            Tensor.repeat(self.ln_final_bias, cur, ctx=ctx0),
            ctx=ctx0,
        )

        # Use the embedding from the EOT token
        eot_idx = embd_inp.argmax()
        cur = Tensor.view_2d(
            cur,
            self.embed_dim,
            1,
            cur.tensor.contents.nb[1],
            eot_idx * cur.tensor.contents.nb[1],
            ctx=ctx0,
        )

        cur = Tensor.mul_mat(
            Tensor.cpy(
                Tensor.transpose(self.text_projection, ctx=ctx0),
                Tensor.new_tensor_2d(wtype, self.embed_dim, self.embed_dim, ctx=ctx0),
            ),
            cur,
            ctx=ctx0,
        )
        gf.build_forward_expand(cur)
        gf.compute()
        return cur

    def _image_encoder_compute_forward_memsize(self):
        e_size = 4
        N = self.grid_size * self.grid_size + 1
        ggml_overhead = 256

        mem_size = 0
        mem_size += 256
        mem_size += (
            e_size * self.image_resolution * self.image_resolution * 3 + ggml_overhead
        )  # image
        mem_size += (
            e_size * self.grid_size * self.grid_size * self.vision_width + ggml_overhead
        )  # conv
        mem_size += e_size * self.vision_width * N + ggml_overhead  # concat

        mem_size += (
            e_size * self.vision_width * N + ggml_overhead
        ) * 2  # Copy in visual features
        mem_size += e_size * 8 + 256
        mem_size += 2 * ggml_overhead  # cpy and transpose
        mem_size += e_size * 8 + 256  # ???
        mem_size += (
            e_size * self.vision_width * N + ggml_overhead
        )  # Copy in positional embeddings (new tensor 2d)
        mem_size += (
            e_size * self.vision_width * N + ggml_overhead
        )  # copy visual features: ret

        mem_size += e_size * self.vision_width * N + ggml_overhead  # add

        mem_size += e_size * self.vision_width * N + ggml_overhead  # ln_pre: norm
        mem_size += e_size * self.vision_width * N + ggml_overhead  # ln_pre: repeat
        mem_size += e_size * self.vision_width * N + ggml_overhead  # ln_pre: repeat
        mem_size += e_size * self.vision_width * N + ggml_overhead  # ln_pre: mul
        mem_size += e_size * self.vision_width * N + ggml_overhead  # ln_pre: add

        res_block_mem_size = ResidualAttentionBlock.compute_forward_mem_size(
            N, self.vision_width, self.vision_heads, use_attn_mask=False
        )

        mem_size += res_block_mem_size * self.vision_layers
        mem_size += ggml_overhead  # ln_post: transpose
        mem_size += (e_size * self.vision_width + ggml_overhead) * 3  # ln_post

        mem_size += e_size * self.vision_width * self.embed_dim + ggml_overhead
        mem_size += ggml_overhead  # cpy
        mem_size += 159808  # Compute Overhead ??
        return mem_size

    def _encode_image_internal(self, image):
        wtype = GGML_TYPE(ggml.ggml_ftype_to_ggml_type(ctypes.c_int(0)))

        mem_size = self._image_encoder_compute_forward_memsize()
        mem_buffer = np.empty(mem_size, dtype=np.uint8)
        init_params = InitParams(
            mem_size=mem_size, mem_buffer=mem_buffer.ctypes.data_as(ctypes.c_void_p)
        )
        ctx0 = Context(init_params=init_params)

        gf = CGraph(cgraph=ggml.ggml_cgraph(n_threads=self.n_threads), ctx=ctx0)

        img_tensor = Tensor.new_tensor_4d(
            wtype,
            image.shape[3],
            image.shape[2],
            image.shape[1],
            image.shape[0],
            ctx=ctx0,
        )
        img_tensor.numpy()[:] = image.permute(3, 2, 1, 0)
        cur = Tensor.conv_2d_sk_p0(
            self.visual.visual_conv1_weight, img_tensor, ctx=ctx0
        )

        cur = Tensor.reshape_2d(
            cur,
            cur.shape[0] * cur.shape[1],
            cur.shape[2],
            ctx=ctx0,
        )

        concat = Tensor.new_tensor_2d(wtype, cur.shape[0] + 1, cur.shape[1], ctx=ctx0)

        concat = Tensor.set_1d(
            concat,
            Tensor.view_1d(
                self.visual.visual_class_embedding,
                self.visual.visual_class_embedding.shape[0],
                0,
            ),
            0,
            ctx=ctx0,
        )

        # Copy in the visual features
        concat = Tensor.set_2d(
            concat,
            Tensor.cpy(
                Tensor.transpose(cur, ctx=ctx0),
                Tensor.new_tensor_2d(wtype, cur.shape[0], cur.shape[1]),
                ctx=ctx0,
            ),
            cur.tensor.contents.nb[1],
            self.visual.visual_class_embedding.nbytes(),
            ctx=ctx0,
        )

        # Copy in the positional embeddings
        cur = Tensor.cpy(
            concat,
            Tensor.new_tensor_2d(wtype, concat.shape[1], concat.shape[0], ctx=ctx0),
            ctx=ctx0,
        )

        pEmb = self.visual.visual_positional_embedding

        cur = Tensor.add(cur, pEmb, ctx=ctx0)

        # ln_pre
        cur = Tensor.norm(cur, ctx=ctx0)

        cur = Tensor.add(
            Tensor.mul(
                Tensor.repeat(self.visual.visual_ln_pre_weight, cur, ctx=ctx0),
                cur,
                ctx=ctx0,
            ),
            Tensor.repeat(self.visual.visual_ln_pre_bias, cur, ctx=ctx0),
            ctx=ctx0,
        )

        # Transformer
        for il in range(self.visual.layers):
            resblock = self.visual.resblocks[il]
            cur = resblock.forward(cur, ctx=ctx0, gf=gf)

        # ln_post
        cur = Tensor.norm(
            Tensor.view_2d(Tensor.transpose(cur, ctx=ctx0), cur.shape[0], 1, 1, 0),
            ctx=ctx0,
        )

        cur = Tensor.add(
            Tensor.mul(
                Tensor.repeat(self.visual.visual_ln_post_weight, cur, ctx=ctx0),
                cur,
                ctx=ctx0,
            ),
            Tensor.repeat(self.visual.visual_ln_post_bias, cur, ctx=ctx0),
            ctx=ctx0,
        )

        # Token Projection
        cur = Tensor.mul_mat(
            Tensor.cpy(
                Tensor.transpose(self.visual.visual_proj),
                Tensor.new_tensor_2d(
                    wtype,
                    self.visual.visual_proj.shape[1],
                    self.visual.visual_proj.shape[0],
                    ctx=ctx0,
                ),
                ctx=ctx0,
            ),
            Tensor.reshape_2d(cur, cur.shape[0], 1),
            ctx=ctx0,
        )

        gf.build_forward_expand(cur)
        gf.compute()

        return cur

    @staticmethod
    def init_from_file(model_file: str, verbose=True, n_threads=1):
        with open(model_file, "rb") as fin:
            # Magic Number
            (magic,) = struct.unpack("i", (fin.read(struct.calcsize("i"))))

            assert magic == ggml.GGML_FILE_MAGIC
            if verbose:
                print("magic number =", hex(magic))
            # Hyperparameters
            (
                vision_width,
                vision_layers,
                vision_patch_size,
                grid_size,
                image_resolution,
                embed_dim,
                context_length,
                transformer_width,
                transformer_heads,
                transformer_layers,
                ftype,
                vocab_size,
            ) = struct.unpack("iiiiiiiiiiii", fin.read(struct.calcsize("iiiiiiiiiiii")))

            qntvr = ftype // ggml.GGML_QNT_VERSION_FACTOR
            if verbose:
                print("vision_width    =", vision_width)
                print("vision_layers    =", vision_layers)
                print("vision_patch_size =", vision_patch_size)
                print("grid_size       =", grid_size)
                print("image_resolution =", image_resolution)
                print("embed_dim       =", embed_dim)
                print("context_length  =", context_length)
                print("transformer_width =", transformer_width)
                print("transformer_heads =", transformer_heads)
                print("transformer_layers =", transformer_layers)
                print("ftype           =", ftype)
                print("qntvr           =", qntvr)
                print("vocab_size      =", vocab_size)
            ftype %= ggml.GGML_QNT_VERSION_FACTOR
            ftype = GGML_FTYPE(int(ftype))

            # Vocabulary
            vocab: List[Tuple[int, str]] = []
            for i in range(vocab_size):
                (s_len,) = struct.unpack("i", fin.read(struct.calcsize("i")))
                s = fin.read(s_len).decode("utf-8")
                vocab.append((i, s))

            # Model Weights
            wtype = GGML_TYPE(ggml.ggml_ftype_to_ggml_type(ctypes.c_int(ftype.value)))

            ctx_size = compute_ctx_size(fin)

            mem_buffer = np.empty(ctx_size, dtype=np.uint8)
            init_params = InitParams(
                mem_size=ctx_size,
                mem_buffer=mem_buffer.ctypes.data_as(ctypes.c_void_p),
            )
            ctx = Context(init_params=init_params)

            # Create Model
            model = ClipModel(
                ctx=ctx,
                wtype=wtype,
                vision_width=vision_width,
                vision_layers=vision_layers,
                vision_patch_size=vision_patch_size,
                image_resolution=image_resolution,
                embed_dim=embed_dim,
                context_length=context_length,
                vocab_size=vocab_size,
                transformer_width=transformer_width,
                transformer_heads=transformer_heads,
                transformer_layers=transformer_layers,
                n_threads=n_threads,
            )

            # Load Weights
            while True:
                nbytes = struct.calcsize("iii")
                data = fin.read(nbytes)
                if len(data) != nbytes:
                    break
                (n_dims, s_len, ftype) = struct.unpack("iii", data)
                dims = struct.unpack(
                    "i" * n_dims, fin.read(struct.calcsize("i" * n_dims))
                )
                tensor_name = fin.read(s_len).decode("utf-8")
                tensor = model.tensors[tensor_name]
                n_elements = tensor.nelements()
                expected_n_elements = np.prod(dims)
                if n_elements != expected_n_elements:
                    raise ValueError(
                        f"tensor {tensor_name} has {n_elements} elements, but {expected_n_elements} were expected"
                    )

                buf = (ctypes.c_char * tensor.nbytes()).from_address(tensor.data)
                offset = fin.tell()
                fname = fin.name.encode("utf-8")
                fin.readinto(buf)

            return model


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-m", "--model", type=str, default=None)
    parser.add_argument("--use-gpu", action="store_true")
    args = parser.parse_args()

    model_file = args.model
    model = ClipModel.init_from_file(model_file, n_threads=1, use_gpu=args.use_gpu)
    image = np.random.rand(3, 224, 224).astype(np.float32)
    output = model.eval([image, image])
    print(output)



"""ggml-python implemention of the Replit code model

Model is available at:
https://huggingface.co/replit/replit-code-v1-3b

This implementation is based on the example model code and ggml model file format from:
https://github.com/ggerganov/ggml/tree/master/examples/replit
"""
from __future__ import annotations
import abc
import math
import struct
import ctypes
import argparse
import multiprocessing
from collections import deque

from typing import (
    Callable,
    Deque,
    Iterator,
    List,
    Optional,
    Sequence,
    Tuple,
    Dict,
    Union,
)

import numpy as np
import numpy.typing as npt

import ggml

from ggml.utils import to_numpy

from contextlib import ExitStack


class ReplitAbortException(Exception):
    pass

## Generic Sampling Functions


def sample(
    logits: npt.NDArray[np.float32],
    last_tokens: Optional[List[int]] = None,
    presence_penalty: float = 0.0,
    frequency_penalty: float = 0.0,
    temperature: float = 1.0,
    top_p: float = 0.0,
) -> int:
    if last_tokens is None:
        last_tokens = []
    if temperature == 0.0:
        return int(np.argmax(logits))
    logits = frequency_and_presence_penalties(
        logits, last_tokens, frequency_penalty, presence_penalty
    )
    return nucleus_sampling(logits, top_p=top_p, temperature=temperature)


# TODO: this is likely incorrect
def frequency_and_presence_penalties(
    logits: npt.NDArray[np.float32],
    last_tokens: Sequence[int],
    alpha_frequency: float,
    alpha_presence: float,
):
    if len(last_tokens) == 0:
        return logits

    if alpha_frequency == 0.0 and alpha_presence == 0.0:
        return logits

    # Calculate the frequency penalty contribution
    frequency_penalty = alpha_frequency * np.log(np.unique(last_tokens).size + 1)

    # Calculate the presence penalty contribution
    presence_penalty = alpha_presence * np.array(
        [float(token in last_tokens) for token in range(len(logits))]
    )

    # Apply penalties to the logits
    penalized_logits = logits - frequency_penalty - presence_penalty

    return penalized_logits


def nucleus_sampling(
    logits: npt.NDArray[np.float32], top_p: float, temperature: float = 1.0
):
    # Apply temperature to logits
    logits /= temperature

    # Subtract the maximum value for numerical stability
    logits -= logits.max()  # type: ignore

    # Calculate probabilities using softmax function with epsilon
    epsilon = 1e-8
    probabilities = np.exp(logits) / ((np.exp(logits)).sum() + epsilon)  # type: ignore

    # Filter out NaN values from probabilities
    probabilities = np.nan_to_num(probabilities)

    # Sort the probabilities in descending order and get the corresponding indices
    sorted_indices = np.argsort(probabilities)[::-1]

    # Select the indices within the nucleus
    nucleus_indices = sorted_indices[: int(len(sorted_indices) * top_p)]

    # Calculate the updated probabilities within the nucleus
    nucleus_probabilities = probabilities[nucleus_indices]

    # Normalize the probabilities within the nucleus
    nucleus_probabilities /= nucleus_probabilities.sum()  # type: ignore

    # Sample from the updated probabilities
    selected_token = np.random.choice(nucleus_indices, p=nucleus_probabilities)

    return selected_token


### Context Buffer


class ContextBuffer(abc.ABC):
    @abc.abstractmethod
    def resize(self, new_size: int) -> None:
        raise NotImplementedError

    @property
    @abc.abstractmethod
    def buffer(self) -> ctypes.c_void_p:
        raise NotImplementedError


class CpuContextBuffer(ContextBuffer):
    def __init__(self, buffer_size: int = 256 * 1024 * 1024):
        self.buffer_size = buffer_size
        self._buffer = (ctypes.c_uint8 * self.buffer_size)()

    def resize(self, new_size: int):
        assert new_size > self.buffer_size

        self.buffer_size = new_size
        ctypes.resize(self._buffer, self.buffer_size)

    @property
    def buffer(self) -> ctypes.c_void_p:
        return ctypes.c_void_p(ctypes.addressof(self._buffer))


### Tokenizer


class Tokenizer(abc.ABC):
    @abc.abstractmethod
    def tokenize(self, text: str) -> List[int]:
        raise NotImplementedError

    @abc.abstractmethod
    def detokenize(self, tokens: List[int]) -> str:
        raise NotImplementedError


class ReplitTokenizer(Tokenizer):
    def __init__(self, vocab: List[Tuple[int, str, float]]):
        self.vocab = vocab
        self.piece_map = {piece: (i, -score) for i, piece, score in self.vocab}
        self.ws_symbol = b"\342\226\201"

    def tokenize(self, text: str) -> List[int]:
        normalized_text = text.replace(" ", self.ws_symbol.decode("utf-8"))
        tokenized, _ = ReplitTokenizer.encode_word(normalized_text, self.piece_map)
        return tokenized

    def detokenize(self, tokens: List[int]) -> str:
        text = "".join(self.vocab[token][1] for token in tokens)
        detokenized = text.replace(self.ws_symbol.decode("utf-8"), " ")
        return detokenized

    @staticmethod
    def encode_word(
        word: str, model: Dict[str, Tuple[int, float]]
    ) -> Tuple[List[int], float]:
        len_word = len(word)
        best_segmentation_starts = [-1] * (len_word + 1)
        best_segmentation_scores = [math.inf] * (len_word + 1)
        best_segmentation_starts[0], best_segmentation_scores[0] = 0, 0.0

        for idx in range(len_word):
            if best_segmentation_starts[idx] != -1:
                end_idx = idx + 1
                while end_idx <= len_word:
                    token = word[idx:end_idx]
                    if token in model:
                        token_score = model[token][1]
                        if (
                            best_segmentation_scores[idx] + token_score
                            < best_segmentation_scores[end_idx]
                        ):
                            best_segmentation_starts[end_idx] = idx
                            best_segmentation_scores[end_idx] = (
                                best_segmentation_scores[idx] + token_score
                            )
                    end_idx += 1

        if best_segmentation_scores[-1] == math.inf:
            return [], 0.0

        tokens: Deque[int] = deque()
        idx = len_word
        while idx > 0:
            start_idx = best_segmentation_starts[idx]
            token = word[start_idx:idx]
            token_id = model[token][0]
            tokens.appendleft(token_id)
            idx = start_idx

        return list(tokens), best_segmentation_scores[-1]


class ReplitSentencepieceTokenizer(Tokenizer):
    def __init__(self, model_path: str):
        import sentencepiece

        self.tokenizer = sentencepiece.SentencePieceProcessor(model_file=model_path)

    def tokenize(self, text: str) -> List[int]:
        return self.tokenizer.encode(text)

    def detokenize(self, tokens: List[int]) -> str:
        return self.tokenizer.decode(tokens)


### Replit Model Definition


class ReplitLayer:
    def __init__(self, wtype: int, n_embd: int, ctx: ggml.ggml_context_p):
        self.norm_1_weight = ggml.ggml_new_tensor_1d(ctx, ggml.GGML_TYPE_F32, n_embd)
        self.c_attn_wqkv_weight = ggml.ggml_new_tensor_2d(
            ctx, wtype, n_embd, 3 * n_embd
        )
        self.c_attn_out_proj_weight = ggml.ggml_new_tensor_2d(
            ctx, wtype, n_embd, n_embd
        )
        self.norm_2_weight = ggml.ggml_new_tensor_1d(ctx, ggml.GGML_TYPE_F32, n_embd)
        self.c_ffn_up_proj_weight = ggml.ggml_new_tensor_2d(
            ctx, wtype, n_embd, 4 * n_embd
        )
        self.c_ffn_down_proj_weight = ggml.ggml_new_tensor_2d(
            ctx, wtype, 4 * n_embd, n_embd
        )


class ReplitModel:
    def __init__(
        self,
        d_model: int,
        max_seq_len: int,
        n_heads: int,
        n_layers: int,
        vocab_size: int,
        ftype: int,
        vocab: List[Tuple[int, str, float]],
        tokenizer: Tokenizer,
        n_batch: int,
        n_threads: int,
        weights_buffer: ContextBuffer,
        ctx: ggml.ggml_context_p,
        cancel_callback: Optional[Callable[[], bool]] = None,
    ):
        self.d_model = d_model
        self.max_seq_len = max_seq_len
        self.n_heads = n_heads
        self.n_layers = n_layers
        self.vocab_size = vocab_size
        self.ftype = ftype
        self.ctx = ctx
        self.layers: List[ReplitLayer] = []
        self.tensors: Dict[str, ggml.ggml_tensor_p] = {}
        self.vocab = vocab
        self.tokenizer = tokenizer
        self.n_batch = n_batch
        self.n_threads = n_threads
        self.weights_buffer = weights_buffer
        self.cancel_callback = cancel_callback

        n_layer = self.n_layers
        n_embd = self.d_model
        n_ctx = self.max_seq_len
        n_vocab = self.vocab_size
        wtype = ggml.ggml_ftype_to_ggml_type(ftype)

        n_mem = n_layer * n_ctx
        n_elements = n_embd * n_mem

        self.memory_k = ggml.ggml_new_tensor_1d(ctx, ggml.GGML_TYPE_F16, n_elements)
        self.memory_v = ggml.ggml_new_tensor_1d(ctx, ggml.GGML_TYPE_F16, n_elements)

        self.wte_weight = ggml.ggml_new_tensor_2d(ctx, wtype, n_embd, n_vocab)
        self.norm_f_weight = ggml.ggml_new_tensor_1d(ctx, ggml.GGML_TYPE_F32, n_embd)
        self.tensors["transformer.wte.weight"] = self.wte_weight
        self.tensors["transformer.norm_f.weight"] = self.norm_f_weight

        self.mem_per_token = 0
        self.eval_buffer = CpuContextBuffer()

        for i in range(n_layer):
            layer = ReplitLayer(
                wtype=wtype,
                n_embd=n_embd,
                ctx=ctx,
            )
            self.layers.append(layer)

            self.tensors[f"transformer.blocks.{i}.norm_1.weight"] = layer.norm_1_weight
            self.tensors[
                f"transformer.blocks.{i}.attn.Wqkv.weight"
            ] = layer.c_attn_wqkv_weight
            self.tensors[
                f"transformer.blocks.{i}.attn.out_proj.weight"
            ] = layer.c_attn_out_proj_weight
            self.tensors[f"transformer.blocks.{i}.norm_2.weight"] = layer.norm_2_weight
            self.tensors[
                f"transformer.blocks.{i}.ffn.up_proj.weight"
            ] = layer.c_ffn_up_proj_weight
            self.tensors[
                f"transformer.blocks.{i}.ffn.down_proj.weight"
            ] = layer.c_ffn_down_proj_weight

        self.n_tokens = 0
        self.input_ids: npt.NDArray[np.intc] = np.ndarray(
            (self.max_seq_len), dtype=np.intc
        )
        self.scores: npt.NDArray[np.single] = np.ndarray(
            (self.max_seq_len, n_vocab), dtype=np.single
        )

    def __del__(self):
        ggml.ggml_free(self.ctx)

    @staticmethod
    def encode_word(
        word: str, model: Dict[str, Tuple[int, float]]
    ) -> Tuple[List[int], float]:
        len_word = len(word)
        best_segmentation_starts = [-1] * (len_word + 1)
        best_segmentation_scores = [math.inf] * (len_word + 1)
        best_segmentation_starts[0], best_segmentation_scores[0] = 0, 0.0

        for idx in range(len_word):
            if best_segmentation_starts[idx] != -1:
                end_idx = idx + 1
                while end_idx <= len_word:
                    token = word[idx:end_idx]
                    if token in model:
                        token_score = model[token][1]
                        if (
                            best_segmentation_scores[idx] + token_score
                            < best_segmentation_scores[end_idx]
                        ):
                            best_segmentation_starts[end_idx] = idx
                            best_segmentation_scores[end_idx] = (
                                best_segmentation_scores[idx] + token_score
                            )
                    end_idx += 1

        if best_segmentation_scores[-1] == math.inf:
            return [], 0.0

        tokens: Deque[int] = deque()
        idx = len_word
        while idx > 0:
            start_idx = best_segmentation_starts[idx]
            token = word[start_idx:idx]
            token_id = model[token][0]
            tokens.appendleft(token_id)
            idx = start_idx

        return list(tokens), best_segmentation_scores[-1]

    def tokenize(self, text: str) -> List[int]:
        return self.tokenizer.tokenize(text)

    def detokenize(self, tokens: List[int]) -> str:
        return self.tokenizer.detokenize(tokens)

    def reset(self):
        self.n_tokens = 0

    def _build_forward(
        self,
        ctx0: ggml.ggml_context_p,
        n_tokens: int,
        n_past: int,
        n_threads: int,
    ):
        N = n_tokens
        n_embd = self.d_model
        n_layer = self.n_layers
        n_ctx = self.max_seq_len
        n_head = self.n_heads

        gf = ggml.ggml_cgraph(n_threads=n_threads)

        embd = ggml.ggml_new_tensor_1d(
            ctx0,
            ggml.GGML_TYPE_I32,
            N,
        )
        ggml.ggml_set_name(embd, b"embd")

        inpL = ggml.ggml_get_rows(ctx0, self.wte_weight, embd)

        for il in range(n_layer):
            # // lctx.use_buf(ctx0, 0)

            # // a = self.ln_1(x)
            cur = ggml.ggml_norm(ctx0, inpL, 1e-5)
            # offload_func(cur)
            ggml.ggml_set_name(cur, b"norm_0")
            cur = ggml.ggml_mul(
                ctx0,
                ggml.ggml_repeat(ctx0, self.layers[il].norm_1_weight, cur),
                cur,
            )
            ggml.ggml_set_name(cur, b"attention_norm_0")

            # // self-attention
            # //  b, _, past_key_value = self.attn(a, past_key_value=past_key_value,
            # //  attn_bias=attn_bias, attention_mask=attention_mask,
            # //  is_causal=is_causal)

            # // compute QKV
            cur = ggml.ggml_mul_mat(ctx0, self.layers[il].c_attn_wqkv_weight, cur)
            ggml.ggml_set_name(cur, b"tmpkqv")

            Qcur = ggml.ggml_view_2d(
                ctx0,
                cur,
                n_embd,
                N,
                cur.contents.nb[1],
                0 * ctypes.sizeof(ctypes.c_float) * n_embd,
            )
            ggml.ggml_set_name(Qcur, b"Qcur")
            Kcur = ggml.ggml_view_2d(
                ctx0,
                cur,
                n_embd,
                N,
                cur.contents.nb[1],
                1 * ctypes.sizeof(ctypes.c_float) * n_embd,
            )
            ggml.ggml_set_name(Kcur, b"Kcur")
            Vcur = ggml.ggml_view_2d(
                ctx0,
                cur,
                n_embd,
                N,
                cur.contents.nb[1],
                2 * ctypes.sizeof(ctypes.c_float) * n_embd,
            )
            ggml.ggml_set_name(Vcur, b"Vcur")

            # // store key and value to memory
            k = ggml.ggml_view_1d(
                ctx0,
                self.memory_k,
                N * n_embd,
                (ggml.ggml_element_size(self.memory_k) * n_embd)
                * (il * n_ctx + n_past),
            )
            ggml.ggml_set_name(k, b"k")
            v = ggml.ggml_view_1d(
                ctx0,
                self.memory_v,
                N * n_embd,
                (ggml.ggml_element_size(self.memory_v) * n_embd)
                * (il * n_ctx + n_past),
            )
            ggml.ggml_set_name(v, b"v")

            ggml.ggml_build_forward_expand(
                ctypes.pointer(gf),
                ggml.ggml_cpy(
                    ctx0,
                    Kcur,
                    k,
                ),
            )
            ggml.ggml_build_forward_expand(
                ctypes.pointer(gf),
                ggml.ggml_cpy(
                    ctx0,
                    Vcur,
                    v,
                ),
            )

            # // Q = Qcur.contiguous().view(n_embd/n_head, n_head, N).permute(0,
            # // 2, 1, 3) [64, N, 12]
            Q = ggml.ggml_permute(
                ctx0,
                ggml.ggml_cpy(
                    ctx0,
                    Qcur,
                    ggml.ggml_new_tensor_3d(
                        ctx0,
                        ggml.GGML_TYPE_F32,
                        n_embd // n_head,
                        n_head,
                        N,
                    ),
                ),
                0,
                2,
                1,
                3,
            )
            ggml.ggml_set_name(Q, b"Q")

            # // K = Kmem.view(n_embd/n_head, n_head, n_past + N).permute(0, 2, 1,
            # // 3) [64, n_past + N, 12]
            K = ggml.ggml_permute(
                ctx0,
                ggml.ggml_reshape_3d(
                    ctx0,
                    ggml.ggml_view_1d(
                        ctx0,
                        self.memory_k,
                        (n_past + N) * n_embd,
                        il * n_ctx * ggml.ggml_element_size(self.memory_k) * n_embd,
                    ),
                    n_embd // n_head,
                    n_head,
                    n_past + N,
                ),
                0,
                2,
                1,
                3,
            )
            ggml.ggml_set_name(K, b"K")

            # // K * Q
            KQ = ggml.ggml_mul_mat(ctx0, K, Q)
            ggml.ggml_set_name(KQ, b"KQ")

            # // KQ_scaled = KQ / sqrt(n_embd/n_head)
            KQ_scaled = ggml.ggml_scale(
                ctx0,
                KQ,
                1.0 / np.sqrt(float(n_embd) / n_head),
            )
            ggml.ggml_set_name(KQ_scaled, b"KQ_scaled")

            KQ_scaled_alibi = ggml.ggml_alibi(
                ctx0,
                KQ_scaled,
                n_past,
                n_head,
                8.0,
            )
            ggml.ggml_set_name(KQ_scaled_alibi, b"KQ_scaled_alibi")

            # // KQ_masked = mask_past(KQ_scaled)
            KQ_masked = ggml.ggml_diag_mask_inf(
                ctx0,
                KQ_scaled_alibi,
                n_past,
            )
            ggml.ggml_set_name(KQ_masked, b"KQ_masked")

            # // KQ = soft_max(KQ_masked)
            KQ_soft_max = ggml.ggml_soft_max(
                ctx0,
                KQ_masked,
            )
            ggml.ggml_set_name(KQ_soft_max, b"KQ_soft_max")

            # // V_trans = Vmem.view(n_embd/n_head, n_head, n_past + N).permute(1,
            # // 2, 0, 3).contiguous() [n_past + N, 64, 12]
            V_trans = ggml.ggml_cpy(
                ctx0,
                ggml.ggml_permute(
                    ctx0,
                    ggml.ggml_reshape_3d(
                        ctx0,
                        ggml.ggml_view_1d(
                            ctx0,
                            self.memory_v,
                            (n_past + N) * n_embd,
                            il * n_ctx * ggml.ggml_element_size(self.memory_v) * n_embd,
                        ),
                        n_embd // n_head,
                        n_head,
                        n_past + N,
                    ),
                    1,
                    2,
                    0,
                    3,
                ),
                ggml.ggml_new_tensor_3d(
                    ctx0,
                    self.memory_v.contents.type,
                    n_past + N,
                    n_embd // n_head,
                    n_head,
                ),
            )
            # offload_func_v(V_trans)
            ggml.ggml_set_name(V_trans, b"V_trans")

            # // KQV = transpose(V) * KQ_soft_max
            KQV = ggml.ggml_mul_mat(ctx0, V_trans, KQ_soft_max)
            # offload_func_v(KQV)
            ggml.ggml_set_name(KQV, b"KQV")

            # // KQV_merged = KQV.permute(0, 2, 1, 3)
            KQV_merged = ggml.ggml_permute(
                ctx0,
                KQV,
                0,
                2,
                1,
                3,
            )
            ggml.ggml_set_name(KQV_merged, b"KQV_merged")

            # // cur = KQV_merged.contiguous().view(n_embd, N)
            cur = ggml.ggml_cpy(
                ctx0,
                KQV_merged,
                ggml.ggml_new_tensor_2d(
                    ctx0,
                    ggml.GGML_TYPE_F32,
                    n_embd,
                    N,
                ),
            )
            ggml.ggml_set_name(cur, b"KQV_merged_contiguous")

            # // projection
            cur = ggml.ggml_mul_mat(
                ctx0,
                self.layers[il].c_attn_out_proj_weight,
                cur,
            )
            ggml.ggml_set_name(cur, b"result_wo")

            # // lctx.use_buf(ctx0, 1)

            inpL = ggml.ggml_add(
                ctx0,
                inpL,
                cur,
            )
            ggml.ggml_set_name(cur, b"inpFF")

            # // m = self.ln_2(x)
            cur = ggml.ggml_norm(ctx0, inpL, 1e-5)
            ggml.ggml_set_name(cur, b"norm_1")
            cur = ggml.ggml_mul(
                ctx0,
                ggml.ggml_repeat(ctx0, self.layers[il].norm_2_weight, cur),
                cur,
            )
            ggml.ggml_set_name(cur, b"norm")

            # // n = self.mlp(m)
            cur = ggml.ggml_mul_mat(
                ctx0,
                self.layers[il].c_ffn_up_proj_weight,
                cur,
            )
            ggml.ggml_set_name(cur, b"result_mlp_up")

            # // GELU activation
            cur = ggml.ggml_gelu(
                ctx0,
                cur,
            )
            ggml.ggml_set_name(cur, b"gelu")
            # // projection
            # // cur = proj_w*cur + proj_b
            cur = ggml.ggml_mul_mat(
                ctx0,
                self.layers[il].c_ffn_down_proj_weight,
                cur,
            )
            ggml.ggml_set_name(cur, b"result_mlp_down")

            # // x = x + n
            inpL = ggml.ggml_add(
                ctx0,
                inpL,
                cur,
            )
            ggml.ggml_set_name(cur, b"inpFF_+_result_mlp_down")

        # // lctx.use_buf(ctx0, 0)

        # // norm
        inpL = ggml.ggml_norm(ctx0, inpL, 1e-5)
        ggml.ggml_set_name(inpL, b"norm_f")

        # // inpL = ln_f_g*inpL
        inpL = ggml.ggml_mul(
            ctx0,
            ggml.ggml_repeat(ctx0, self.norm_f_weight, inpL),
            inpL,
        )
        ggml.ggml_set_name(inpL, b"norm_f_mul")

        # // output embedding weight tied to input embedding
        inpL = ggml.ggml_mul_mat(
            ctx0,
            self.wte_weight,
            inpL,
        )
        ggml.ggml_set_name(inpL, b"result_output")

        # // lctx.use_buf(ctx0, -1)

        ggml.ggml_build_forward_expand(ctypes.pointer(gf), inpL)

        return gf

    def _eval_internal(self, embd_inp: Sequence[int], n_past: int, n_threads: int):
        N = len(embd_inp)
        n_vocab = self.vocab_size
        required_buffer_size = int(self.mem_per_token * N * 2.0)
        if (
            self.mem_per_token > 0
            and self.eval_buffer.buffer_size < required_buffer_size
        ):
            self.eval_buffer.resize(required_buffer_size)

        init_params = ggml.ggml_init_params(
            mem_size=self.eval_buffer.buffer_size,
            mem_buffer=self.eval_buffer.buffer,
            no_alloc=False,
        )
        exit_stack = ExitStack()
        ctx0 = ggml.ggml_init(init_params)
        if ctx0 is None:
            raise RuntimeError("Failed to initialize GGML context")
        exit_stack.callback(ggml.ggml_free, self.ctx)
        gf = self._build_forward(ctx0, len(embd_inp), n_past, n_threads)
        embd = ggml.ggml_graph_get_tensor(ctypes.pointer(gf), b"embd")
        assert embd is not None
        inpL = ggml.ggml_graph_get_tensor(ctypes.pointer(gf), b"result_output")
        assert inpL is not None
        to_numpy(embd)[:] = np.array(embd_inp, dtype=np.int32)
        gp = ggml.ggml_graph_plan(ctypes.pointer(gf), self.n_threads)
        work_data = (ctypes.c_uint8 * gp.work_size)()
        gp.work_data = ctypes.cast(work_data, ctypes.POINTER(ctypes.c_uint8))
        if self.cancel_callback is not None:

            @ggml.ggml_abort_callback
            def abort_callback(data: ctypes.c_void_p) -> Union[ctypes.c_bool, bool]:
                assert self.cancel_callback is not None
                return self.cancel_callback()

            self._abort_callback = abort_callback  # NOTE: keep reference
            gp.abort_callback = abort_callback
        rc = ggml.ggml_graph_compute(ctypes.pointer(gf), ctypes.pointer(gp))
        if rc != ggml.GGML_EXIT_SUCCESS:
            raise ReplitAbortException("Execution aborted")
        embd_w = to_numpy(inpL).reshape(
            -1, n_vocab
        )  # .copy() # NOTE: likely wrong to not copy here
        if self.mem_per_token == 0:
            self.mem_per_token = int(ggml.ggml_used_mem(ctx0) / N)
        return embd_w

    def eval(self, tokens: Sequence[int]):
        if self.mem_per_token == 0:
            try:
                self._eval_internal([1, 2, 3, 4], n_past=0, n_threads=self.n_threads)
            except ReplitAbortException as e:
                self.n_tokens = 0
                raise e
        n_ctx = self.max_seq_len
        for i in range(0, len(tokens), self.n_batch):
            batch = tokens[i : min(len(tokens), i + self.n_batch)]
            n_past = min(n_ctx - len(batch), self.n_tokens)
            try:
                scores = self._eval_internal(
                    batch,
                    n_past,
                    self.n_threads,
                )
                # Save tokens
                self.input_ids[self.n_tokens : self.n_tokens + len(batch)] = batch
                # Save logits
                self.scores[self.n_tokens : self.n_tokens + len(batch), :] = scores
                # Update token count
                self.n_tokens += len(batch)
            except ReplitAbortException as e:
                self.n_tokens = n_past
                raise e
        return self.scores[: self.n_tokens, :]

    def generate(
        self,
        tokens: Sequence[int],
        top_p: float = 0.95,
        temperature: float = 0.80,
        frequency_penalty: float = 0.0,
        presence_penalty: float = 0.0,
    ) -> Iterator[int]:
        reset = True
        if self.n_tokens > 0:
            longest_prefix = 0
            for a, b in zip(self.input_ids[: self.n_tokens], tokens[:-1]):
                if a == b:
                    longest_prefix += 1
                else:
                    break
            if longest_prefix > 0:
                reset = False
                tokens = tokens[longest_prefix:]
                self.n_tokens = longest_prefix

        if reset:
            self.reset()

        while True:
            scores = self.eval(tokens)
            logits = scores[-1, :]
            token = sample(
                logits,
                top_p=top_p,
                temperature=temperature,
                frequency_penalty=frequency_penalty,
                presence_penalty=presence_penalty,
            )
            yield token
            tokens = [token]

    @staticmethod
    def eos_token():
        return 1

    @staticmethod
    def init_from_file(
        model_file: str,
        n_gpu_layers: int = 0,
        n_batch: int = 1,
        n_threads: int = 1,
        tokenizer: Optional[Tokenizer] = None,
        verbose: bool = True,
        cancel_callback: Optional[Callable[[], bool]] = None,
    ) -> ReplitModel:
        with open(model_file, "rb") as fin:
            # Magic Number
            (magic,) = struct.unpack("i", (fin.read(struct.calcsize("i"))))
            assert magic == ggml.GGML_FILE_MAGIC
            if verbose:
                print("magic number =", hex(magic))
            # Hyperparameters
            d_model, max_seq_len, n_heads, n_layers, vocab_size, ftype = struct.unpack(
                "iiiiii", (fin.read(struct.calcsize("iiiiii")))
            )
            qntvr = ftype // ggml.GGML_QNT_VERSION_FACTOR
            if verbose:
                print("d_model      =", d_model)
                print("max_seq_len  =", max_seq_len)
                print("n_heads      =", n_heads)
                print("n_layers     =", n_layers)
                print("vocab_size   =", vocab_size)
                print("ftype        =", ftype)
                print("qntvr        =", qntvr)
            ftype %= ggml.GGML_QNT_VERSION_FACTOR
            # Vocabulary
            vocab: List[Tuple[int, str, float]] = []
            for i in range(vocab_size):
                (s_len,) = struct.unpack("i", (fin.read(struct.calcsize("i"))))
                s = fin.read(s_len).decode("utf-8")
                (score,) = struct.unpack("f", (fin.read(struct.calcsize("f"))))
                vocab.append((i, s, score))
            # Model Weights
            wtype = ggml.ggml_ftype_to_ggml_type(ftype)

            n_embd = d_model
            n_layer = n_layers
            n_ctx = max_seq_len
            n_vocab = vocab_size

            ctx_size = ReplitModel.compute_ctx_size(
                n_embd=n_embd,
                n_layer=n_layer,
                n_ctx=n_ctx,
                n_vocab=n_vocab,
                wtype=wtype,
            )

            if verbose:
                print("ctx size     =", ctx_size // (1024 * 1024), "MB")

            # create context
            weights_buffer = CpuContextBuffer(ctx_size)
            init_params = ggml.ggml_init_params(
                mem_size=ctx_size,
                mem_buffer=weights_buffer.buffer,
                no_alloc=False,
            )
            ctx = ggml.ggml_init(init_params)
            if ctx is None:
                raise RuntimeError("Failed to initialize GGML context")

            model = ReplitModel(
                # hyperparameters
                d_model=d_model,
                max_seq_len=max_seq_len,
                n_heads=n_heads,
                n_layers=n_layers,
                vocab_size=vocab_size,
                ftype=ftype,
                # vocabulary
                vocab=vocab,
                tokenizer=ReplitTokenizer(vocab) if tokenizer is None else tokenizer,
                ctx=ctx,
                n_batch=n_batch,
                n_threads=n_threads,
                weights_buffer=weights_buffer,
                # misc
                cancel_callback=cancel_callback,
            )

            n_tensors = 0
            total_size = 0

            while True:
                nbytes = struct.calcsize("iii")
                data = fin.read(nbytes)
                if len(data) != nbytes:
                    break
                n_dims, length, ttype = struct.unpack("iii", data)
                nelements = 1
                ne = [1, 1]
                for i in range(n_dims):
                    (dim,) = struct.unpack("i", (fin.read(struct.calcsize("i"))))
                    ne[i] = dim
                    nelements *= ne[i]
                name = fin.read(length).decode("utf-8")
                if name not in model.tensors:
                    raise ValueError(f"Tensor {name} not found in model")
                tensor = model.tensors[name]
                if ggml.ggml_nelements(tensor) != nelements:
                    raise ValueError(
                        f"Tensor {name} has {ggml.ggml_nelements(tensor)} elements, but {nelements} expected"
                    )
                if tensor.contents.ne[0] != ne[0] or tensor.contents.ne[1] != ne[1]:
                    raise ValueError(
                        f"Tensor {name} has {tensor.contents.ne[0]}x{tensor.contents.ne[1]} shape, but {ne[0]}x{ne[1]} expected"
                    )
                bpe = ggml.ggml_type_size(ttype)
                if (
                    (nelements * bpe) / ggml.ggml_blck_size(tensor.contents.type)
                ) != ggml.ggml_nbytes(tensor):
                    raise ValueError(
                        f"Tensor {name} has {ggml.ggml_nbytes(tensor)} bytes, but {(nelements * bpe) / ggml.ggml_blck_size(tensor.contents.type)} expected"
                    )
                tensor_data = ggml.ggml_get_data(tensor)
                if tensor_data is None:
                    raise ValueError(f"Failed to get data for tensor {name}")
                fin.readinto(
                    (ctypes.c_uint8 * ggml.ggml_nbytes(tensor)).from_address(
                        tensor_data
                    )
                )

                total_size += ggml.ggml_nbytes(tensor)
                if n_tensors % 8 == 0:
                    print(".", end="", flush=True)
                n_tensors += 1
            print("done")
            print(
                "model size =",
                total_size // (1024 * 1024),
                "MB",
                "num tensors =",
                n_tensors,
            )
        return model

    @staticmethod
    def compute_ctx_size(
        n_embd: int,
        n_layer: int,
        n_ctx: int,
        n_vocab: int,
        wtype: int,
    ) -> int:
        wtype_sizef = ggml.ggml_type_sizef(wtype)
        f32_sizef = ggml.ggml_type_sizef(ggml.GGML_TYPE_F32)
        f16_sizef = ggml.ggml_type_sizef(ggml.GGML_TYPE_F16)

        ctx_size = 0
        ctx_size += n_embd * n_vocab * wtype_sizef
        ctx_size += n_embd * f32_sizef

        ctx_size += n_layer * (n_embd * f32_sizef)
        ctx_size += n_layer * (3 * n_embd * n_embd * wtype_sizef)
        ctx_size += n_layer * (n_embd**2 * wtype_sizef)
        ctx_size += n_layer * (n_embd * f32_sizef)
        ctx_size += n_layer * (4 * n_embd * n_embd * wtype_sizef)
        ctx_size += n_layer * (n_embd**2 * 4 * wtype_sizef)

        ctx_size += n_ctx * n_layer * n_embd * f16_sizef
        ctx_size += n_ctx * n_layer * n_embd * f16_sizef

        ctx_size += (1 + 6 * n_layer) * 512
        ctx_size = int(ctx_size)
        return ctx_size


if __name__ == "__main__":
    parser = argparse.ArgumentParser()
    parser.add_argument("-m", "--model", type=str, default=None)
    parser.add_argument("-p", "--prompt", type=str, default="def fib(n):")
    parser.add_argument(
        "--n_threads", type=int, default=max(1, multiprocessing.cpu_count() // 2)
    )
    parser.add_argument("--n_batch", type=int, default=512)
    parser.add_argument("--n_gpu_layers", type=int, default=0)
    parser.add_argument("--max_tokens", type=int, default=32)
    parser.add_argument("--temperature", type=float, default=1.0)
    parser.add_argument("--top_p", type=float, default=1.0)
    parser.add_argument("--presence_penalty", type=float, default=0.0)
    parser.add_argument("--frequency_penalty", type=float, default=0.0)
    args = parser.parse_args()

    model_file = args.model
    n_threads = args.n_threads
    n_batch = args.n_batch
    n_gpu_layers = args.n_gpu_layers
    max_tokens = args.max_tokens
    temperature = args.temperature
    top_p = args.top_p
    presence_penalty = args.presence_penalty
    frequency_penalty = args.frequency_penalty

    model = ReplitModel.init_from_file(
        model_file, n_gpu_layers=n_gpu_layers, n_threads=n_threads, n_batch=n_batch
    )

    prompt = args.prompt
    prompt_tokens = model.tokenize(prompt)
    all_tokens: List[int] = prompt_tokens[:]  # type: ignore
    n_past = 0
    tokens: List[int] = prompt_tokens[:]  # type: ignore

    print("number of tokens in prompt =", len(prompt_tokens))
    for i, token_id in enumerate(prompt_tokens):
        print(f"token[{i}] =", token_id)

    print()
    print(prompt, end="", flush=True)
    for _ in range(max_tokens):
        # eval
        scores = model.eval(tokens)
        logits = scores[-1, :]
        # sample
        token_id = sample(
            logits,
            last_tokens=all_tokens,
            temperature=temperature,
            top_p=top_p,
            presence_penalty=presence_penalty,
            frequency_penalty=frequency_penalty,
        )
        if token_id == model.eos_token():
            break
        # update
        all_tokens.append(token_id)
        print(model.detokenize([token_id]), end="", flush=True)
        n_past += len(tokens)
        tokens = [token_id]
    print()



# Generates bindings for the ggml library.
#
# cffi requires prior C preprocessing of the headers, and it uses pycparser which chokes on a couple of things
# so we help it a bit (e.g. replace sizeof expressions with their value, remove exotic syntax found in Darwin headers).
import os, sys, re, subprocess
import cffi
from stubs import generate_stubs

API = os.environ.get('API', 'api.h')
CC = os.environ.get('CC') or 'gcc'
C_INCLUDE_DIR = os.environ.get('C_INCLUDE_DIR', '../../../llama.cpp')
CPPFLAGS = [
    "-I", C_INCLUDE_DIR,
    '-D__fp16=uint16_t',  # pycparser doesn't support __fp16
    '-D__attribute__(x)=',
    '-D_Static_assert(x, m)=',
] + [x for x in os.environ.get('CPPFLAGS', '').split(' ') if x != '']

try: header = subprocess.run([CC, "-E", *CPPFLAGS, API], capture_output=True, text=True, check=True).stdout
except subprocess.CalledProcessError as e: print(f'{e.stderr}\n{e}', file=sys.stderr); raise

header = '\n'.join([l for l in header.split('\n') if '__darwin_va_list' not in l]) # pycparser hates this

# Replace constant size expressions w/ their value (compile & run a mini exe for each, because why not).
# First, extract anyting *inside* square brackets and anything that looks like a sizeof call.
for expr in set(re.findall(f'(?<=\\[)[^\\]]+(?=])|sizeof\\s*\\([^()]+\\)', header)):
    if re.match(r'^(\d+|\s*)$', expr): continue # skip constants and empty bracket contents
    subprocess.run([CC, "-o", "eval_size_expr", *CPPFLAGS, "-x", "c", "-"], text=True, check=True,
                   input=f'''#include <stdio.h>
                             #include "{API}"
                             int main() {{ printf("%lu", (size_t)({expr})); }}''')
    size = subprocess.run(["./eval_size_expr"], capture_output=True, text=True, check=True).stdout
    print(f'Computed constexpr {expr} = {size}')
    header = header.replace(expr, size)

ffibuilder = cffi.FFI()
ffibuilder.cdef(header)
ffibuilder.set_source(f'ggml.cffi', None) # we're not compiling a native extension, as this quickly gets hairy
ffibuilder.compile(verbose=True)

with open("ggml/__init__.pyi", "wt") as f:
    f.write(generate_stubs(header))


import ggml
import ggml.utils
import ctypes
import pytest
import numpy as np

from ggml.utils import setup_sigabrt_handler

setup_sigabrt_handler()

ggml_metal_available = ggml.GGML_USE_METAL

run_if_ggml_metal_available = pytest.mark.skipif(
    not ggml_metal_available,
    reason="METAL not available",
)

@run_if_ggml_metal_available
def test_metal():
    n_tensors = 1 + 2 # input (x) and weights (a, b)
    params = ggml.ggml_init_params(
        mem_size=ggml.ggml_tensor_overhead() * n_tensors, mem_buffer=None, no_alloc=True
    )
    ctx = ggml.ggml_init(params)
    assert ctx is not None

    backend = ggml.ggml_backend_metal_init()

    assert backend is not None

    # create the tensors for input and weights
    x = ggml.ggml_new_tensor_1d(ctx, ggml.GGML_TYPE_F32, 1)

    a = ggml.ggml_new_tensor_1d(ctx, ggml.GGML_TYPE_F32, 1)
    b = ggml.ggml_new_tensor_1d(ctx, ggml.GGML_TYPE_F32, 1)

    # allocate the tensors in the backend
    buffer = ggml.ggml_backend_alloc_ctx_tensors(ctx, backend)

    # set the values of the weights
    ggml.ggml_backend_tensor_set(
        a,
        ctypes.cast(np.array([3.0], dtype=np.single).ctypes.data, ctypes.c_void_p),
        0,
        ggml.ggml_nbytes(a),
    )
    ggml.ggml_backend_tensor_set(
        b,
        ctypes.cast(np.array([4.0], dtype=np.single).ctypes.data, ctypes.c_void_p),
        0,
        ggml.ggml_nbytes(a),
    )

    max_nodes = 4096

    buf_size = (
        ggml.ggml_tensor_overhead() * max_nodes
        + ggml.ggml_graph_overhead_custom(max_nodes, False)
    )
    buf = (ctypes.c_uint8 * buf_size)()

    def build_graph(
        x: ggml.ggml_tensor_p, a: ggml.ggml_tensor_p, b: ggml.ggml_tensor_p
    ):
        params = ggml.ggml_init_params(
            mem_size=buf_size,
            mem_buffer=ctypes.cast(buf, ctypes.c_void_p),
            no_alloc=True,
        )
        ctx0 = ggml.ggml_init(params)

        assert ctx0 is not None

        gf = ggml.ggml_new_graph_custom(ctx0, max_nodes, False)

        x2 = ggml.ggml_mul(ctx0, x, x)
        ax2 = ggml.ggml_mul(ctx0, a, x2)
        f = ggml.ggml_add(ctx0, ax2, b)

        ggml.ggml_set_name(x2, b"x2")
        ggml.ggml_set_name(ax2, b"ax2")
        ggml.ggml_set_name(f, b"f")

        ggml.ggml_build_forward_expand(gf, f)

        ggml.ggml_free(ctx0)

        return gf

    allocr = ggml.ggml_gallocr_new(ggml.ggml_backend_get_default_buffer_type(backend))

    gf = build_graph(x, a, b)

    ggml.ggml_gallocr_reserve(allocr, gf)

    gf = build_graph(x, a, b)

    ggml.ggml_gallocr_alloc_graph(allocr, gf)

    ggml.ggml_backend_tensor_set(
        x,
        ctypes.cast(np.array([2.0], dtype=np.single).ctypes.data, ctypes.c_void_p),
        0,
        ggml.ggml_nbytes(x),
    )

    ggml.ggml_backend_graph_compute(backend, gf)

    f = ggml.ggml_graph_get_tensor(gf, b"f")

    output = np.zeros(1, dtype=np.single)
    ggml.ggml_backend_tensor_get(
        f, ctypes.cast(output.ctypes.data, ctypes.c_void_p), 0, ggml.ggml_nbytes(x)
    )

    assert output[0] == 16.0

    ggml.ggml_gallocr_free(allocr)
    ggml.ggml_backend_buffer_free(buffer)
    ggml.ggml_backend_free(backend)
    ggml.ggml_free(ctx)



import http.server
import socketserver
import os
import sys

DIRECTORY = os.path.abspath(os.path.join(os.path.dirname(__file__), 'web'))
PORT = 8000

class CustomHTTPRequestHandler(http.server.SimpleHTTPRequestHandler):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, directory=DIRECTORY, **kwargs)

    def end_headers(self):
        # Add required headers for SharedArrayBuffer
        self.send_header("Cross-Origin-Opener-Policy", "same-origin")
        self.send_header("Cross-Origin-Embedder-Policy", "require-corp")
        self.send_header("Access-Control-Allow-Origin", "*")
        super().end_headers()

# Enable address reuse
class CustomServer(socketserver.TCPServer):
    allow_reuse_address = True

try:
    with CustomServer(("", PORT), CustomHTTPRequestHandler) as httpd:
        print(f"Serving directory '{DIRECTORY}' at http://localhost:{PORT}")
        print(f"Application context root: http://localhost:{PORT}/")
        try:
            httpd.serve_forever()
        except KeyboardInterrupt:
            print("\nServer stopped.")
            # Force complete exit
            sys.exit(0)
except OSError as e:
    print(f"Error: {e}")
    sys.exit(1)



import ctypes

import ggml
import ggml.utils

import jax

from typing import Optional

params = ggml.ggml_init_params(mem_size=16 * 1024 * 1024, mem_buffer=None)
ctx = ggml.ggml_init(params)
x_in = ggml.ggml_new_tensor_1d(ctx, ggml.GGML_TYPE_F32, 1)

@ggml.ggml_custom1_op_t
def double(
    tensor_out: ggml.ggml_tensor_p,
    tensor_in: ggml.ggml_tensor_p,
    ith: int,
    nth: int,
    userdata: Optional[ctypes.c_void_p],
):
    x = jax.device_put(ggml.utils.to_numpy(tensor_in))
    x *= 2
    ggml.utils.to_numpy(tensor_out)[:] = jax.device_get(x)

x_out = ggml.ggml_map_custom1(ctx, x_in, double, 1, None)
gf = ggml.ggml_build_forward(x_out)

ggml.ggml_set_f32(x_in, 21.0)

ggml.ggml_graph_compute_with_ctx(ctx, ctypes.pointer(gf), 1)
output = ggml.ggml_get_f32_1d(x_out, 0)
assert output == 42.0
print("GGML output: ", output)
ggml.ggml_free(ctx)


"""Utility functions for ggml-python.
"""
from __future__ import annotations

import enum
import ctypes
import signal
import platform
import traceback

from typing import Any, Optional, Sequence, Tuple

from ggml import ggml

import numpy as np
import numpy.typing as npt


class GGML_TYPE(enum.IntEnum):
    F32 = ggml.GGML_TYPE_F32
    F16 = ggml.GGML_TYPE_F16
    Q4_0 = ggml.GGML_TYPE_Q4_0
    Q4_1 = ggml.GGML_TYPE_Q4_1
    Q5_0 = ggml.GGML_TYPE_Q5_0
    Q5_1 = ggml.GGML_TYPE_Q5_1
    Q8_0 = ggml.GGML_TYPE_Q8_0
    Q8_1 = ggml.GGML_TYPE_Q8_1
    I8 = ggml.GGML_TYPE_I8
    I16 = ggml.GGML_TYPE_I16
    I32 = ggml.GGML_TYPE_I32


NUMPY_DTYPE_TO_GGML_TYPE = {
    np.float16: GGML_TYPE.F16,
    np.float32: GGML_TYPE.F32,
    np.int8: GGML_TYPE.I8,
    np.int16: GGML_TYPE.I16,
    np.int32: GGML_TYPE.I32,
}

GGML_TYPE_TO_NUMPY_DTYPE = {v: k for k, v in NUMPY_DTYPE_TO_GGML_TYPE.items()}


def to_numpy(
    tensor: ggml.ggml_tensor_p,
    shape: Optional[Tuple[int, ...]] = None,
) -> npt.NDArray[Any]:
    """Get the data of a ggml tensor as a numpy array.

    Parameters:
        tensor: ggml tensor

    Returns:
        Numpy array with a view of data from tensor
    """
    ggml_type = GGML_TYPE(tensor.contents.type)
    if ggml_type == GGML_TYPE.F16:
        ctypes_type = ctypes.c_uint16
    else:
        ctypes_type = np.ctypeslib.as_ctypes_type(GGML_TYPE_TO_NUMPY_DTYPE[ggml_type])

    data = ggml.ggml_get_data(tensor)
    if data is None:
        raise ValueError("tensor data is None")
    array = (ctypes_type * ggml.ggml_nelements(tensor)).from_address(data)
    n_dims = ggml.ggml_n_dims(tensor)
    shape_ = tuple(reversed(tensor.contents.ne[:n_dims]))
    strides = tuple(reversed(tensor.contents.nb[:n_dims]))
    output = np.ctypeslib.as_array(array)
    if ggml_type == GGML_TYPE.F16:
        output.dtype = np.float16  # type: ignore
    return np.lib.stride_tricks.as_strided(
        output, shape=shape if shape is not None else shape_, strides=strides
    )


def from_numpy(x: npt.NDArray[Any], ctx: ggml.ggml_context_p) -> ggml.ggml_tensor_p:
    """Create a new ggml tensor with data copied from a numpy array.

    Parameters:
        x: numpy array
        ctx: ggml context

    Returns:
        New ggml tensor with data copied from x
    """
    ggml_type = NUMPY_DTYPE_TO_GGML_TYPE[x.dtype.type]
    shape = tuple(reversed(x.shape))
    tensor = ggml.ggml_new_tensor(
        ctx,
        ggml_type.value,
        len(shape),
        (ctypes.c_int64 * len(shape))(*shape),
    )
    tensor.contents.nb[: len(shape)] = (ctypes.c_int64 * len(shape))(
        *tuple(reversed(x.strides))
    )
    if ggml.ggml_get_data(tensor) is not None:
        to_numpy(tensor)[:] = x
    return tensor


def copy_to_cpu(
    ctx: ggml.ggml_context_p, tensor: ggml.ggml_tensor_p
) -> ggml.ggml_tensor_p:
    """Copy a ggml tensor from a GPU backend to CPU.

    Parameters:
        ctx: ggml context
        tensor: ggml tensor

    Returns:
        New ggml tensor with data copied from tensor on CPU backend"""
    tmp = ggml.ggml_dup_tensor(ctx, tensor)
    to_numpy(tmp)[:] = 0
    return ggml.ggml_add_inplace(ctx, tmp, tensor)


def quantize_0(
    data_f32: ggml.CtypesArray[ctypes.c_float],
    nelements: int,
    ne0: int,
    ttype: GGML_TYPE,
    work: Optional[ggml.CtypesArray[ctypes.c_float]] = None,
    imatrix: Optional[ggml.CtypesArray[ctypes.c_float]] = None,
):
    """Quantize a float32 array.

    Parameters:
        data_f32: float32 array
        nelements: number of elements in data_f32
        ne0: number of elements in data_f32 that are zero
        ttype: ggml type to quantize to
        work: work buffer
        imatrix: quantization matrix

    Returns:
        (work, cur_size): outpuut buffer, histogram, number of bytes in work buffer
    """
    work = work or (ctypes.c_float * nelements)()
    cur_size = ggml.ggml_quantize_chunk(
        ttype,
        data_f32,
        ctypes.cast(work, ctypes.c_void_p),
        0,
        nelements,
        ne0,
        imatrix,
    )
    return ctypes.cast(work, ctypes.c_void_p), cur_size


def quantize_row(
    data_f32: ggml.CtypesArray[ctypes.c_float],
    nelements: int,
    ttype: GGML_TYPE,
    work: Optional[ctypes.c_void_p] = None,
):
    """Quantize a row of a ggml tensor.

    Parameters:
        data_f32: float32 array
        nelements: number of elements in data_f32
        ttype: ggml type to quantize to
        work: work buffer

    Returns:
        output buffer"""
    type_traits = ggml.ggml_internal_get_type_traits(ttype.value)
    from_float = type_traits.from_float
    work = work or ctypes.cast((ctypes.c_float * nelements)(), ctypes.c_void_p)
    from_float(data_f32, work, nelements)
    return work


def dequantize_row(
    data_q: ctypes.c_void_p,
    nelements: int,
    ttype: GGML_TYPE,
    work: Optional[ctypes.c_void_p] = None,
):
    """Dequantize a row of a ggml tensor.

    Parameters:
        data_q: quantized data
        nelements: number of elements in data_q
        ttype: ggml type to dequantize from
        work: work buffer

    Returns:
        output buffer"""
    type_traits = ggml.ggml_internal_get_type_traits(ttype.value)
    to_float = type_traits.to_float
    work = work or ctypes.cast((ctypes.c_float * nelements)(), ctypes.c_void_p)
    to_float(data_q, work, nelements)
    return work


def get_ndims(tensor: ggml.ggml_tensor_p) -> int:
    """Get the number of dimensions of a ggml tensor.

    Parameters:
        tensor: ggml tensor

    Returns:
        Number of dimensions of tensor
    """
    return ggml.ggml_n_dims(tensor)


def get_shape(tensor: ggml.ggml_tensor_p) -> Tuple[int, ...]:
    """Get the shape of a ggml tensor.

    Parameters:
        tensor: ggml tensor

    Returns:
        Shape of tensor
    """
    return tuple(tensor.contents.ne[: ggml.ggml_n_dims(tensor)])


def get_strides(tensor: ggml.ggml_tensor_p) -> Tuple[int, ...]:
    """Get the strides of a ggml tensor.

    Parameters:
        tensor: ggml tensor

    Returns:
        Strides of tensor
    """
    return tuple(tensor.contents.nb[: ggml.ggml_n_dims(tensor)])


def slice_tensor(
    ctx: ggml.ggml_context_p, tensor: ggml.ggml_tensor_p, indices: Sequence[slice]
):
    """Slice a ggml tensor along multiple dimensions.

    The slice is a view of the original tensor with the same number of dimensions.

    Parameters:
        ctx: ggml context
        tensor: ggml tensor
        indices: indices to slice along

    Returns:
        New ggml tensor slice view"""
    ndims = ggml.ggml_n_dims(tensor)

    # check that the number of dimensions match
    if len(indices) != ndims:
        raise ValueError(
            f"tensor has {ndims} dimensions but {len(indices)} indices were given"
        )

    # calculate slice
    start = tuple(idx.start or 0 for idx in indices)
    end = tuple(idx.stop or get_shape(tensor)[i] for i, idx in enumerate(indices))
    step = tuple(idx.step or 1 for idx in indices)

    # get the shape of the slice
    shape = tuple((end[i] - start[i] + step[i] - 1) // step[i] for i in range(ndims))

    # get the strides of the slice
    strides = tuple(get_strides(tensor)[i] * step[i] for i in range(ndims))

    # get the offset of the slice
    offset = sum(get_strides(tensor)[i] * start[i] for i in range(ndims))

    if ndims == 1:
        return ggml.ggml_view_1d(
            ctx,
            tensor,
            shape[0],
            offset,
        )
    elif ndims == 2:
        return ggml.ggml_view_2d(
            ctx,
            tensor,
            shape[0],
            shape[1],
            strides[1],
            offset,
        )
    elif ndims == 3:
        return ggml.ggml_view_3d(
            ctx,
            tensor,
            shape[0],
            shape[1],
            shape[2],
            strides[1],
            strides[2],
            offset,
        )
    elif ndims == 4:
        return ggml.ggml_view_4d(
            ctx,
            tensor,
            shape[0],
            shape[1],
            shape[2],
            shape[3],
            strides[1],
            strides[2],
            strides[3],
            offset,
        )
    else:
        raise NotImplementedError(
            f"ggml tensors with {ndims} dimensions are not supported"
        )


def setup_sigabrt_handler():
    if platform.system() == "Windows":
        return

    c_globals = ctypes.CDLL(None)  # POSIX
    signal_type = signal.SIGABRT

    @ctypes.CFUNCTYPE(None, ctypes.c_int)
    def sigabrt_handler(sig):  # type: ignore
        traceback.print_stack()
        raise Exception("GGML SIGABRT")

    c_globals.signal(signal_type, sigabrt_handler)



# Convert a model checkpoint to a ggml compatible file
#
# Load the model using TensorFlow.
# Iterate over all variables and write them to a binary file.
#
# For each variable, write the following:
#   - Number of dimensions (int)
#   - Name length (int)
#   - Dimensions (int[n_dims])
#   - Name (char[name_length])
#   - Data (float[n_dims])
#
# By default, the bigger matrices are converted to 16-bit floats.
# This can be disabled by adding the "use-f32" CLI argument.
#
# At the start of the ggml file we write the model parameters
# and vocabulary.
#

import sys
import json
import struct
import numpy as np
import tensorflow as tf

# ref: https://github.com/openai/gpt-2/blob/master/src/encoder.py
def bytes_to_unicode():
    """
    Returns list of utf-8 byte and a corresponding list of unicode strings.
    The reversible bpe codes work on unicode strings.
    This means you need a large # of unicode characters in your vocab if you want to avoid UNKs.
    When you're at something like a 10B token dataset you end up needing around 5K for decent coverage.
    This is a signficant percentage of your normal, say, 32K bpe vocab.
    To avoid that, we want lookup tables between utf-8 bytes and unicode strings.
    And avoids mapping to whitespace/control characters the bpe code barfs on.
    """
    bs = list(range(ord("!"), ord("~")+1))+list(range(ord("¡"), ord("¬")+1))+list(range(ord("®"), ord("ÿ")+1))
    cs = bs[:]
    n = 0
    for b in range(2**8):
        if b not in bs:
            bs.append(b)
            cs.append(2**8+n)
            n += 1
    cs = [chr(n) for n in cs]
    return dict(zip(bs, cs))

# helper method to convert a numpy array to different float types
def convert_to_ftype(data, ftype):
    # fp16
    if ftype == 1:
        return data.astype(np.float16)

    assert False, "Invalid ftype: " + str(ftype)

if len(sys.argv) < 3:
    print("Usage: convert-ckpt-to-ggml.py dir-model ftype\n")
    print("  ftype == 0 -> float32")
    print("  ftype == 1 -> float16")
    sys.exit(1)

# output in the same directory as the model
dir_model = sys.argv[1]
fname_out = sys.argv[1] + "/ggml-model.bin"

with open(dir_model + "/encoder.json", "r", encoding="utf-8") as f:
    encoder = json.load(f)

with open(dir_model + "/hparams.json", "r", encoding="utf-8") as f:
    hparams = json.load(f)

# possible data types
#   ftype == 0 -> float32
#   ftype == 1 -> float16
#
# map from ftype to string
ftype_str = ["f32", "f16"]

ftype = 1
if len(sys.argv) > 2:
    ftype = int(sys.argv[2])
    if ftype < 0 or ftype > 1:
        print("Invalid ftype: " + str(ftype))
        sys.exit(1)
    fname_out = sys.argv[1] + "/ggml-model-" + ftype_str[ftype] + ".bin"

list_vars = tf.train.list_variables(dir_model)

fout = open(fname_out, "wb")

fout.write(struct.pack("i", 0x67676d6c)) # magic: ggml in hex
fout.write(struct.pack("i", hparams["n_vocab"]))
fout.write(struct.pack("i", hparams["n_ctx"]))
fout.write(struct.pack("i", hparams["n_embd"]))
fout.write(struct.pack("i", hparams["n_head"]))
fout.write(struct.pack("i", hparams["n_layer"]))
fout.write(struct.pack("i", ftype))

byte_encoder = bytes_to_unicode()
byte_decoder = {v:k for k, v in byte_encoder.items()}

fout.write(struct.pack("i", len(encoder)))

for key in encoder:
    text = bytearray([byte_decoder[c] for c in key])
    fout.write(struct.pack("i", len(text)))
    fout.write(text)

for name, shape in list_vars:
    print("Processing variable: " + name + " with shape: ", shape)

    data = tf.train.load_variable(dir_model, name).squeeze()
    n_dims = len(data.shape);

    # for efficiency - transpose the projection matrices
    # "model/h.*/attn/c_attn/w"
    # "model/h.*/attn/c_proj/w"
    # "model/h.*/mlp/c_fc/w"
    # "model/h.*/mlp/c_proj/w"
    if name[-14:] == "/attn/c_attn/w" or \
       name[-14:] == "/attn/c_proj/w" or \
       name[-11:] == "/mlp/c_fc/w" or \
       name[-13:] == "/mlp/c_proj/w":
        print("  Transposing")
        data = data.transpose()

    dshape = data.shape

    ftype_cur = 0
    if ftype != 0:
        # match name:
        #  "model/wte"
        #  "model/h.*/attn/c_attn/w"
        #  "model/h.*/attn/c_proj/w"
        #  "model/h.*/mlp/c_fc/w"
        #  "model/h.*/mlp/c_proj/w"
        if name == "model/wte" or name[-2:] == "/w":
            print("  Converting to " + ftype_str[ftype])
            data = convert_to_ftype(data, ftype)
            ftype_cur = ftype
        else:
            print("  Converting to float32")
            data = data.astype(np.float32)
            ftype_cur = 0

    # header
    str = name.encode('utf-8')
    fout.write(struct.pack("iii", n_dims, len(str), ftype_cur))
    for i in range(n_dims):
        fout.write(struct.pack("i", dshape[n_dims - 1 - i]))
    fout.write(str);

    # data
    data.tofile(fout)

fout.close()

print("Done. Output file: " + fname_out)
print("")



"""
  Common helpers for working with ggml + numpy
"""
from ggml import ffi, lib
from typing import Union, Optional
import numpy as np

def init(mem_size: int, mem_buffer: ffi.CData = ffi.NULL, no_alloc: bool = False) -> ffi.CData:
    """
      Initialize a ggml context, which will be freed automatically when the pointer is garbage collected.
    """
    params = ffi.new('struct ggml_init_params*')
    params.mem_size = mem_size
    params.mem_buffer = mem_buffer
    params.no_alloc = no_alloc
    return ffi.gc(lib.ggml_init(params[0]), lib.ggml_free)
 
TensorLike = Union[ffi.CData, np.ndarray]

def copy(from_tensor: TensorLike, to_tensor: TensorLike, allow_requantize: bool = True):
    """
      Copy the contents of one tensor to another, doing any necessary (de/re)quantization transparently.
      Works across numpy & ggml tensors, but they must have the same shape (and be contiguous).

      Parameters
      ----------
      from_tensor : TensorLike
          The tensor to copy from (a numpy array or possibly-quantized ggml tensor)
      to_tensor : TensorLike
          The tensor to copy to (a numpy array or possibly-quantized ggml tensor)
      allow_requantize : bool
          If False, will throw an error if requantization is required (i.e. both from_tensor
          and to_tensor are quantized with different quantization types)
    """
    if id(from_tensor) == id(to_tensor):
        return
 
    __expect_same_layout("source", from_tensor, "destination", to_tensor)
    __check_shape_consistent_with_type(from_tensor)
    __check_shape_consistent_with_type(to_tensor)

    from_type = __get_type(from_tensor)
    to_type = __get_type(to_tensor)

    if from_type == to_type:
        ffi.memmove(__get_data(to_tensor), __get_data(from_tensor), __get_nbytes(from_tensor))
    else:
        assert allow_requantize or not lib.ggml_is_quantized(from_type) or not lib.ggml_is_quantized(to_type), \
            f"Requantizing from {__type_name(from_type)} to {__type_name(to_type)} is disabled. Force with allow_requantize=True"
 
        __set_floats(to_tensor, __get_floats(from_tensor))

def numpy(tensor: ffi.CData, allow_copy: Union[bool, np.ndarray] = False, allow_requantize=False) -> np.ndarray:
    """
      Convert a ggml tensor to a numpy array.
      If the tensor isn't quantized, the returned numpy array will be a view over its data.
 
      If it is quantized (and allow_copy is True), the copy will involve dequantization and the returned array will
      be a copy of the original tensor (any changes to the numpy array won't then be reflected back to the tensor).

      Parameters
      ----------
      tensor : ffi.CData
          The tensor to convert to a numpy array
      allow_copy : bool or np.ndarray
          If False, will throw an error if the tensor is quantized (since dequantization requires extra memory).
          If True, will dequantize the tensor and return a copy of the data in a new float32 numpy array.
          If an np.ndarray, will copy the data into the given array (which must be the same shape as the tensor) when dequantization is needed
      allow_requantize : bool
          If allow_copy is a tensor with a different quantization type than the source tensor, will throw an error unless allow_requantize is True.
    """
    shape = __get_shape(tensor)

    if lib.ggml_is_quantized(tensor.type):
        if allow_copy == False:
            raise ValueError(f"{__describe(tensor)} is quantized, conversion to numpy requires a copy (pass allow_copy=True; changes to the numpy array won't affect the original).")
        elif isinstance(allow_copy, np.ndarray):
            __expect_same_layout("source tensor", tensor, "dequantization output tensor", allow_copy)
            destination = allow_copy
        else:
            destination = np.empty(shape, dtype=np.float32)

        copy(tensor, destination, allow_requantize=allow_requantize)
        return destination
    else:
        dtype = __type_to_dtype(tensor.type)
        if not dtype:
            raise NotImplementedError(f'Cannot convert {__describe(tensor)} to numpy')

        assert __is_contiguous(tensor), f"Cannot convert {__describe(tensor)} to numpy (support contiguous tensors only)"
        nbytes = lib.ggml_nelements(tensor) * lib.ggml_type_size(tensor.type)
        array = np.frombuffer(ffi.buffer(lib.ggml_get_data(tensor), nbytes), dtype=dtype)
        array.shape = shape
        return array

def __type_name(type: int) -> str:
    name = lib.ggml_type_name(type)
    return ffi.string(name).decode('utf-8') if name else None

__k_quant_types = set([
  lib.GGML_TYPE_Q2_K,
  lib.GGML_TYPE_Q3_K,
  lib.GGML_TYPE_Q4_K,
  lib.GGML_TYPE_Q5_K,
  lib.GGML_TYPE_Q6_K,
  lib.GGML_TYPE_Q8_K,
])

__type_to_dtype_dict = {
  lib.GGML_TYPE_I8: np.int8,
  lib.GGML_TYPE_I16: np.int16,
  lib.GGML_TYPE_I32: np.int32,
  lib.GGML_TYPE_F16: np.float16,
  lib.GGML_TYPE_F32: np.float32,
}

def __type_to_dtype(type: int) -> Optional[np.dtype]: return __type_to_dtype_dict.get(type)
def __dtype_to_type(dtype: np.dtype):
    if dtype == np.float32: return lib.GGML_TYPE_F32
    elif dtype == np.float16: return lib.GGML_TYPE_F16
    elif dtype == np.int32: return lib.GGML_TYPE_I32
    elif dtype == np.int16: return lib.GGML_TYPE_I16
    elif dtype == np.int8: return lib.GGML_TYPE_I8
    else: raise ValueError(f"Unsupported dtype: {dtype}")

def __describe(tensor: ffi.CType): return f'Tensor[{__type_name(__get_type(tensor))}, {__get_shape(tensor)}]'
def __get_type(tensor: TensorLike): return __dtype_to_type(tensor.dtype) if isinstance(tensor, np.ndarray) else tensor.type
def __get_shape(x: TensorLike): return x.shape if isinstance(x, np.ndarray) else tuple([x.ne[i] for i in range(x.n_dims)])
def __get_strides(x: TensorLike): return x.strides if isinstance(x, np.ndarray) else tuple([x.nb[i] for i in range(x.n_dims)])
def __get_data(x: TensorLike) -> ffi.CData: return ffi.from_buffer(x) if isinstance(x, np.ndarray) else lib.ggml_get_data(x)
def __get_nbytes(tensor: TensorLike): return tensor.nbytes if isinstance(tensor, np.ndarray) else lib.ggml_nbytes(tensor)
def __get_nelements(tensor: TensorLike): return tensor.size if isinstance(tensor, np.ndarray) else lib.ggml_nelements(tensor)
def __is_contiguous(tensor: TensorLike): return tensor.flags['C_CONTIGUOUS'] if isinstance(tensor, np.ndarray) else lib.ggml_is_contiguous(tensor)

def __get_floats(tensor: TensorLike) -> ffi.CData:
    data, type = __get_data(tensor), __get_type(tensor)
    if type == lib.GGML_TYPE_F32:
        return ffi.cast('float*', data)
    else:
      nelements = __get_nelements(tensor)
      floats = ffi.new('float[]', nelements)
      if type == lib.GGML_TYPE_F16:
          lib.ggml_fp16_to_fp32_row(ffi.cast('uint16_t*', data), floats, nelements)
      elif lib.ggml_is_quantized(type):
          qtype = lib.ggml_internal_get_type_traits(type)
          assert qtype.to_float, f"Type {__type_name(type)} is not supported by ggml"
          qtype.to_float(data, floats, nelements)
      else:
          raise NotImplementedError(f'Cannot read floats from {__describe(tensor)}')
      return floats

def __set_floats(tensor: TensorLike, f32_data: ffi.CData) -> None:
    data, type, nbytes = __get_data(tensor), __get_type(tensor), __get_nbytes(tensor)
    if type == lib.GGML_TYPE_F32:
        ffi.memmove(data, f32_data, nbytes)
    else:
      nelements = __get_nelements(tensor)
      if type == lib.GGML_TYPE_F16:
          lib.ggml_fp32_to_fp16_row(f32_data, ffi.cast('uint16_t*', data), nelements)
      elif lib.ggml_is_quantized(type):
          qtype = lib.ggml_internal_get_type_traits(type)
          assert qtype.from_float, f"Type {__type_name(type)} is not supported by ggml"
          qtype.from_float(f32_data, data, nelements)
      else:
          raise NotImplementedError(f'Cannot write floats to {__describe(tensor)}')

def __expect_same_layout(name1: str, tensor1: TensorLike, name2: str, tensor2: TensorLike):
    shape1, shape2 = __get_shape(tensor1), __get_shape(tensor2)
    assert shape1 == shape2, f"Shape mismatch: {name1} has {shape1} but {name2} has {shape2}"
    assert __is_contiguous(tensor1) and __is_contiguous(tensor2), f"Only contiguous tensors are supported (got {name1} with strides {__get_strides(tensor1)} and {name2} with strides {__get_strides(tensor2)})"

def __check_shape_consistent_with_type(tensor: TensorLike):
    type = __get_type(tensor)
    if not lib.ggml_is_quantized(type):
        return
    shape = __get_shape(tensor)

    block_size = lib.ggml_blck_size(type)
    assert not (block_size == 0 and type in __k_quant_types), f"Can't quantize, native library was not compiled with USE_K_QUANTS!"
    assert block_size > 0, f"Invalid block size {block_size} for type {__type_name(type)}"
    for i, d in enumerate(shape):
        assert d % block_size == 0, f"Dimension {i} of {__describe(tensor)} is not divisible by {block_size}, required for quantization."



# Convert a SAM model checkpoint to a ggml compatible file
#

import sys
import torch
import struct
import numpy as np

if len(sys.argv) < 3:
    print("Usage: convert-pth-to-ggml.py file-model dir-output [ftype]\n")
    print("  ftype == 0 -> float32")
    print("  ftype == 1 -> float16")
    sys.exit(1)

# output in the same directory as the model
fname_model = sys.argv[1]
dir_out     = sys.argv[2]
fname_out   = dir_out + "/ggml-model.bin"

# possible data types
#   ftype == 0 -> float32
#   ftype == 1 -> float16
#
# map from ftype to string
ftype_str = ["f32", "f16"]

ftype = 1
if len(sys.argv) > 3:
    ftype = int(sys.argv[3])

if ftype < 0 or ftype > 1:
    print("Invalid ftype: " + str(ftype))
    sys.exit(1)

fname_out = fname_out.replace(".bin", "-" + ftype_str[ftype] + ".bin")

# Default params are set to sam_vit_b checkpoint
n_enc_state = 768
n_enc_layers = 12
n_enc_heads = 12
n_enc_out_chans = 256
n_pt_embd = 4

model = torch.load(fname_model, map_location="cpu")
for k, v in model.items():
    print(k, v.shape)
    if k == "image_encoder.blocks.0.norm1.weight":
        n_enc_state = v.shape[0]

if n_enc_state == 1024: # sam_vit_l
    n_enc_layers = 24
    n_enc_heads  = 16
elif n_enc_state == 1280: # sam_vit_h
    n_enc_layers = 32
    n_enc_heads  = 16

hparams = {
    "n_enc_state":      n_enc_state,
    "n_enc_layers":     n_enc_layers,
    "n_enc_heads":      n_enc_heads,
    "n_enc_out_chans":  n_enc_out_chans,
    "n_pt_embd":        n_pt_embd,
}

print(hparams)

for k, v in model.items():
    print(k, v.shape)

#exit()
#code.interact(local=locals())

fout = open(fname_out, "wb")

fout.write(struct.pack("i", 0x67676d6c)) # magic: ggml in hex
fout.write(struct.pack("i", hparams["n_enc_state"]))
fout.write(struct.pack("i", hparams["n_enc_layers"]))
fout.write(struct.pack("i", hparams["n_enc_heads"]))
fout.write(struct.pack("i", hparams["n_enc_out_chans"]))
fout.write(struct.pack("i", hparams["n_pt_embd"]))
fout.write(struct.pack("i", ftype))

for k, v in model.items():
    name = k
    shape = v.shape

    if name[:19] == "prompt_encoder.mask":
        continue

    print("Processing variable: " + name + " with shape: ", shape, " and type: ", v.dtype)

    #data = tf.train.load_variable(dir_model, name).squeeze()
    #data = v.numpy().squeeze()
    data = v.numpy()
    n_dims = len(data.shape)

    # for efficiency - transpose some matrices
    # "model/h.*/attn/c_attn/w"
    # "model/h.*/attn/c_proj/w"
    # "model/h.*/mlp/c_fc/w"
    # "model/h.*/mlp/c_proj/w"
    #if name[-14:] == "/attn/c_attn/w" or \
    #   name[-14:] == "/attn/c_proj/w" or \
    #   name[-11:] == "/mlp/c_fc/w" or \
    #   name[-13:] == "/mlp/c_proj/w":
    #    print("  Transposing")
    #    data = data.transpose()

    dshape = data.shape

    # default type is fp16
    ftype_cur = 1
    if ftype == 0 or n_dims == 1 or \
            name == "image_encoder.pos_embed" or \
            name.startswith("prompt_encoder") or \
            name.startswith("mask_decoder.iou_token") or \
            name.startswith("mask_decoder.mask_tokens"):
        print("  Converting to float32")
        data = data.astype(np.float32)
        ftype_cur = 0
    else:
        print("  Converting to float16")
        data = data.astype(np.float16)

    # reshape the 1D bias into a 4D tensor so we can use ggml_repeat
    # keep it in F32 since the data is small
    if name == "image_encoder.patch_embed.proj.bias":
        data = data.reshape(1, data.shape[0], 1, 1)
        n_dims = len(data.shape)
        dshape = data.shape

    print("  New shape: ", dshape)

    # header
    str = name.encode('utf-8')
    fout.write(struct.pack("iii", n_dims, len(str), ftype_cur))
    for i in range(n_dims):
        fout.write(struct.pack("i", dshape[n_dims - 1 - i]))
    fout.write(str)

    # data
    data.tofile(fout)

fout.close()

print("Done. Output file: " + fname_out)
print("")



# %% [markdown]
"""
# Single-batch stochastic gradient descent example using ggml

This example demonstrates how to use ggml to implement a simple SGD optimizer.
"""
# %%
import ggml
import random

a_real = 3.0
b_real = 4.0

ctx0 = ggml.ggml_init(ggml.ggml_init_params(
    mem_size=128 * 1024 * 1024, mem_buffer=None, no_alloc=False
))

assert ctx0 is not None

# define parameters
a = ggml.ggml_new_tensor_1d(ctx0, ggml.GGML_TYPE_F32, 1)
ggml.ggml_set_param(ctx0, a)

b = ggml.ggml_new_tensor_1d(ctx0, ggml.GGML_TYPE_F32, 1)
ggml.ggml_set_param(ctx0, b)

# define input and output
x = ggml.ggml_new_tensor_1d(ctx0, ggml.GGML_TYPE_F32, 1)
ggml.ggml_set_input(x)

tmp = ggml.ggml_mul(ctx0, a, x)
f = ggml.ggml_add(ctx0, tmp, b)

# define loss
f_true = ggml.ggml_new_tensor_1d(ctx0, ggml.GGML_TYPE_F32, 1)
ggml.ggml_set_input(f_true)

tmp = ggml.ggml_sub(ctx0, f, f_true)
loss = ggml.ggml_mul(ctx0, tmp, tmp)

# build forward and backward graph
gf = ggml.ggml_new_graph_custom(ctx0, ggml.GGML_DEFAULT_GRAPH_SIZE, True)
ggml.ggml_build_forward_expand(gf, loss)
gb = ggml.ggml_graph_dup(ctx0, gf)
ggml.ggml_build_backward_expand(ctx0, gf, gb, False)

# initialize parameters
ggml.ggml_set_f32(a, 1.0)
ggml.ggml_set_f32(b, 1.0)

# SGD
lr = 1e-2
nsteps = 1000
decay = 1e-3

for i in range(nsteps):
    # sample data
    x_sample = random.uniform(-10, 10)
    f_sample = a_real * x_sample + b_real

    # set input
    ggml.ggml_set_f32(x, x_sample)
    ggml.ggml_set_f32(f_true, f_sample)

    # reset graph
    ggml.ggml_graph_reset(gf)
    ggml.ggml_set_f32(loss.contents.grad, 1.0)

    # compute forward and backward
    ggml.ggml_graph_compute_with_ctx(ctx0, gb, 1)

    # print loss
    loss_ = ggml.ggml_get_f32_1d(loss, 0)
    print(f"step {i}: loss = {loss_}")

    # decay learning rate
    lr *= (1.0 - decay)

    # update parameters
    ggml.ggml_set_f32(a, ggml.ggml_get_f32_1d(a, 0) - lr * ggml.ggml_get_f32_1d(a.contents.grad, 0))
    ggml.ggml_set_f32(b, ggml.ggml_get_f32_1d(b, 0) - lr * ggml.ggml_get_f32_1d(b.contents.grad, 0))

    # print parameters
    print(f"a = {ggml.ggml_get_f32_1d(a, 0):.2f}, b = {ggml.ggml_get_f32_1d(b, 0):.2f}")


ggml.ggml_free(ctx0)

# %%



# Convert CLIP model from PyTorch to ggml format
#
# Usage: python convert-pt-to-ggml.py ViT-B-32 ./models
#
# This script loads the specified model and clip assets and saves them in ggml format.
# The output is a single binary file containing the following information:
#
#  - hparams
#  - tokenizer vocab
#  - model variables
#
# For each variable, write the following:
#
#  - Number of dimensions (int)
#  - Name length (int)
#  - Dimensions (int[n_dims])
#  - Name (char[name_length])
#  - Data (float[n_dims])
#

import os
import sys
import struct
import gzip
import numpy as np
import clip

if len(sys.argv) < 3:
    print("Usage: convert-pt-to-ggml.py clip_model dir-output\n")
    sys.exit(1)

clip_model = sys.argv[1]
dir_out = sys.argv[2]

# CLIP repo needs to exist at the root directory
MODELS = clip.clip._MODELS
model_filename = os.path.basename(MODELS[clip_model]).replace(".pt", "")

model = clip.load(clip_model, device="cpu")
state_dict = model[0].state_dict()

# output in the same directory as the model
fname_out = os.path.join(dir_out, model_filename + ".ggml")
os.makedirs(dir_out, exist_ok=True)

fout = open(fname_out, "wb")

# Get HParams
# Only ViT models supported for now
vit = True
if vit:
    vision_width = state_dict["visual.conv1.weight"].shape[0]
    vision_layers = len(
        [
            k
            for k in state_dict.keys()
            if k.startswith("visual.") and k.endswith(".attn.in_proj_weight")
        ]
    )
    vision_patch_size = state_dict["visual.conv1.weight"].shape[-1]
    grid_size = round((state_dict["visual.positional_embedding"].shape[0] - 1) ** 0.5)
    image_resolution = vision_patch_size * grid_size
embed_dim = state_dict["text_projection"].shape[1]
context_length = state_dict["positional_embedding"].shape[0]
vocab_size = state_dict["token_embedding.weight"].shape[0]
transformer_width = state_dict["ln_final.weight"].shape[0]
transformer_heads = transformer_width // 64
transformer_layers = len(
    set(k.split(".")[2] for k in state_dict if k.startswith("transformer.resblocks"))
)
print("HParams:")
print("  vision_width:", vision_width)
print("  vision_layers:", vision_layers)
print("  vision_patch_size:", vision_patch_size)
print("  grid_size:", grid_size)
print("  image_resolution:", image_resolution)
print("  embed_dim:", embed_dim)
print("  context_length:", context_length)
print("  vocab_size:", vocab_size)
print("  transformer_width:", transformer_width)
print("  transformer_heads:", transformer_heads)
print("  transformer_layers:", transformer_layers)


ftype = 0

# Write hparams
fout.write(struct.pack("i", 0x67676D6C))  # magic: ggml in hex
fout.write(struct.pack("i", vision_width))
fout.write(struct.pack("i", vision_layers))
fout.write(struct.pack("i", vision_patch_size))
fout.write(struct.pack("i", grid_size))
fout.write(struct.pack("i", image_resolution))
fout.write(struct.pack("i", embed_dim))
fout.write(struct.pack("i", context_length))
fout.write(struct.pack("i", transformer_width))
fout.write(struct.pack("i", transformer_heads))
fout.write(struct.pack("i", transformer_layers))
fout.write(struct.pack("i", ftype))  # ftype: 0 = float32, 1 = float16

bpe_path = os.path.join(os.path.dirname(clip.__file__), "bpe_simple_vocab_16e6.txt.gz")
merges = gzip.open(bpe_path).read().decode("utf-8").split("\n")
merges = merges[1 : 49152 - 256 - 2 + 1]
merges = [tuple(merge.split()) for merge in merges]

vocab = list(clip.simple_tokenizer.bytes_to_unicode().values())
tokens = vocab + [v + "</w>" for v in vocab]
for merge in merges:
    tokens.append("".join(merge))
tokens.extend(["<|startoftext|>", "<|endoftext|>"])
# byte_decoder = {v: k for k, v in clip.simple_tokenizer.bytes_to_unicode().items()}

fout.write(struct.pack("i", len(tokens)))

for key in tokens:
    text = key.encode("utf-8")
    fout.write(struct.pack("i", len(text)))
    fout.write(text)

for name in state_dict.keys():
    data = state_dict[name].squeeze().numpy()
    print("Processing variable: " + name + " with shape: ", data.shape)
    # ftype == 0 -> float32, ftype == 1 -> float16
    ftype = 0

    if name == "visual.conv1.weight":
        data = data.astype(np.float16)
        ftype = 1
    n_dims = len(data.shape)

    # header
    str = name.encode("utf-8")
    fout.write(struct.pack("iii", n_dims, len(str), ftype))
    for i in range(n_dims):
        fout.write(struct.pack("i", data.shape[n_dims - 1 - i]))
    fout.write(str)

    # data
    data.tofile(fout)

fout.close()

print("Done. Output file: " + fname_out)
print("")



from __future__ import annotations

import time
import uuid
import json
import multiprocessing
from functools import partial
from threading import Lock
from typing import (
    Callable,
    Dict,
    List,
    Optional,
    Union,
    Iterator,
    AsyncIterator,
    Sequence,
)
from os import environ

from typing_extensions import TypedDict, Literal

import numpy as np
import numpy.typing as npt

import anyio
from anyio.streams.memory import MemoryObjectSendStream
from starlette.concurrency import run_in_threadpool, iterate_in_threadpool
from fastapi import FastAPI, Request, Depends
from pydantic import BaseModel, BaseSettings, Field, create_model_from_typeddict
from sse_starlette.sse import EventSourceResponse

from main import ReplitModel, ReplitSentencepieceTokenizer


## Types
class CompletionLogprobs(TypedDict):
    text_offset: List[int]
    token_logprobs: List[Optional[float]]
    tokens: List[str]
    top_logprobs: List[Optional[Dict[str, float]]]


class CompletionChoice(TypedDict):
    text: str
    index: int
    logprobs: Optional[CompletionLogprobs]
    finish_reason: Optional[str]


class CompletionUsage(TypedDict):
    prompt_tokens: int
    completion_tokens: int
    total_tokens: int


class CompletionChunk(TypedDict):
    id: str
    object: Literal["text_completion"]
    created: int
    model: str
    choices: List[CompletionChoice]


class Completion(TypedDict):
    id: str
    object: Literal["text_completion"]
    created: int
    model: str
    choices: List[CompletionChoice]
    usage: CompletionUsage


class OpenAIify:
    def __init__(
        self,
        model: ReplitModel,
        cancel_callback: Optional[Callable[[], bool]] = None,
    ):
        self.model = model
        self.cancel_callback = cancel_callback

    def tokenize(self, text: str) -> List[int]:
        return self.model.tokenize(text)

    def detokenize(self, tokens: List[int]) -> str:
        return self.model.detokenize(tokens)

    def generate(
        self,
        tokens: Sequence[int],
        top_p: float = 0.95,
        temperature: float = 0.80,
        frequency_penalty: float = 0.0,
        presence_penalty: float = 0.0,
    ) -> Iterator[int]:
        return self.model.generate(
            tokens,
            top_p=top_p,
            temperature=temperature,
            frequency_penalty=frequency_penalty,
            presence_penalty=presence_penalty,
        )

    def _create_completion(
        self,
        prompt: str,
        suffix: Optional[str] = None,
        max_tokens: int = 16,
        temperature: float = 0.8,
        top_p: float = 0.95,
        logprobs: Optional[int] = None,
        echo: bool = False,
        stop: Optional[Union[str, List[str]]] = [],
        frequency_penalty: float = 0.0,
        presence_penalty: float = 0.0,
        stream: bool = False,
        model: Optional[str] = None,
    ) -> Union[Iterator[Completion], Iterator[CompletionChunk]]:
        completion_id: str = f"cmpl-{str(uuid.uuid4())}"
        created: int = int(time.time())
        completion_tokens: List[int] = []
        # Add blank space to start of prompt to match OG llama tokenizer
        prompt_tokens: List[int] = self.tokenize(prompt)
        text: str = ""
        returned_tokens: int = 0
        stop = (
            stop if isinstance(stop, list) else [stop] if isinstance(stop, str) else []
        )
        model_name: str = model if model is not None else "replit-code-v1-3b"

        # Truncate prompt if it is too long
        max_tokens = min(
            max_tokens, max(0, self.model.max_seq_len - len(prompt_tokens) - 1)
        )
        if len(prompt_tokens) + max_tokens > self.model.max_seq_len:
            raise ValueError(
                f"Requested tokens exceed context window of {self.model.max_seq_len}"
            )

        stop_sequences = stop if stop != [] else []
        finish_reason = "length"
        for token in self.generate(
            prompt_tokens,
            top_p=top_p,
            temperature=temperature,
            frequency_penalty=frequency_penalty,
            presence_penalty=presence_penalty,
        ):
            if token == self.eos_token():
                text = self.detokenize(completion_tokens)
                finish_reason = "stop"
                break

            if self.cancel_callback is not None and self.cancel_callback():
                text = self.detokenize(completion_tokens)
                finish_reason = "stop"
                break

            completion_tokens.append(token)

            all_text = self.detokenize(completion_tokens)
            any_stop = [s for s in stop_sequences if s in all_text]
            if len(any_stop) > 0:
                first_stop = any_stop[0]
                text = all_text[: all_text.index(first_stop)]
                finish_reason = "stop"
                break

            if stream:
                remaining_tokens = completion_tokens[returned_tokens:]
                remaining_text = self.detokenize(remaining_tokens)
                remaining_length = len(remaining_text)

                # We want to avoid yielding any characters from
                # the generated text if they are part of a stop
                # sequence.
                first_stop_position = 0
                for s in stop_sequences:
                    for i in range(min(len(s), remaining_length), 0, -1):
                        if remaining_text.endswith(s[:i]):
                            if i > first_stop_position:
                                first_stop_position = i
                            break

                token_end_position = 0
                for token in remaining_tokens:
                    token_end_position += len(self.detokenize([token]))
                    # Check if stop sequence is in the token
                    if token_end_position >= (
                        remaining_length - first_stop_position - 1
                    ):
                        break
                    logprobs_or_none: Optional[CompletionLogprobs] = None
                    if logprobs is not None:
                        token_str = self.detokenize([token])
                        text_offset = len(prompt) + len(
                            self.detokenize(completion_tokens[:returned_tokens])
                        )
                        token_offset = len(prompt_tokens) + returned_tokens
                        logits = self.model.scores[token_offset - 1, :].tolist()
                        current_logprobs = self.logits_to_logprobs(logits)
                        sorted_logprobs = list(
                            sorted(
                                zip(current_logprobs, range(len(current_logprobs))),
                                reverse=True,
                            )
                        )
                        top_logprob = {
                            self.detokenize([i]): logprob
                            for logprob, i in sorted_logprobs[:logprobs]
                        }
                        top_logprob[token_str] = current_logprobs[int(token)]
                        logprobs_or_none = {
                            "tokens": [self.detokenize([token])],
                            "text_offset": [text_offset],
                            "token_logprobs": [sorted_logprobs[int(token)][0]],
                            "top_logprobs": [top_logprob],
                        }
                    returned_tokens += 1
                    yield {
                        "id": completion_id,
                        "object": "text_completion",
                        "created": created,
                        "model": model_name,
                        "choices": [
                            {
                                "text": self.detokenize([token]),
                                "index": 0,
                                "logprobs": logprobs_or_none,
                                "finish_reason": None,
                            }
                        ],
                    }

            if len(completion_tokens) >= max_tokens:
                text = self.detokenize(completion_tokens)
                finish_reason = "length"
                break

        if stream:
            remaining_tokens = completion_tokens[returned_tokens:]
            all_text = self.detokenize(remaining_tokens)
            any_stop = [s for s in stop_sequences if s in all_text]
            if len(any_stop) > 0:
                end = min(all_text.index(stop) for stop in any_stop)
            else:
                end = len(all_text)

            token_end_position = 0
            for token in remaining_tokens:
                token_end_position += len(self.detokenize([token]))

                logprobs_or_none: Optional[CompletionLogprobs] = None
                if logprobs is not None:
                    token_str = self.detokenize([token])
                    text_offset = len(prompt) + len(
                        self.detokenize(completion_tokens[:returned_tokens])
                    )
                    token_offset = len(prompt_tokens) + returned_tokens - 1
                    logits = self.model.scores[token_offset, :].tolist()
                    current_logprobs = self.logits_to_logprobs(logits)
                    sorted_logprobs = list(
                        sorted(
                            zip(current_logprobs, range(len(current_logprobs))),
                            reverse=True,
                        )
                    )
                    top_logprob = {
                        self.detokenize([i]): logprob
                        for logprob, i in sorted_logprobs[:logprobs]
                    }
                    top_logprob[token_str] = current_logprobs[int(token)]
                    logprobs_or_none = {
                        "tokens": [self.detokenize([token])],
                        "text_offset": [text_offset],
                        "token_logprobs": [sorted_logprobs[int(token)][0]],
                        "top_logprobs": [top_logprob],
                    }

                if token_end_position >= end:
                    last_text = self.detokenize([token])
                    if token_end_position == end - 1:
                        break
                    returned_tokens += 1
                    yield {
                        "id": completion_id,
                        "object": "text_completion",
                        "created": created,
                        "model": model_name,
                        "choices": [
                            {
                                "text": last_text[
                                    : len(last_text) - (token_end_position - end)
                                ],
                                "index": 0,
                                "logprobs": logprobs_or_none,
                                "finish_reason": finish_reason,
                            }
                        ],
                    }
                    break
                returned_tokens += 1
                yield {
                    "id": completion_id,
                    "object": "text_completion",
                    "created": created,
                    "model": model_name,
                    "choices": [
                        {
                            "text": self.detokenize([token]),
                            "index": 0,
                            "logprobs": logprobs_or_none,
                            "finish_reason": finish_reason
                            if returned_tokens == len(completion_tokens)
                            else None,
                        }
                    ],
                }
            return

        text_str = text

        if echo:
            text_str = prompt + text_str

        if suffix is not None:
            text_str = text_str + suffix

        logprobs_or_none: Optional[CompletionLogprobs] = None
        if logprobs is not None:
            text_offset = 0 if echo else len(prompt)
            token_offset = 0 if echo else len(prompt_tokens[1:])
            text_offsets: List[int] = []
            token_logprobs: List[Optional[float]] = []
            tokens: List[str] = []
            top_logprobs: List[Optional[Dict[str, float]]] = []

            if echo:
                # Remove leading BOS token
                all_tokens = prompt_tokens[1:] + completion_tokens
            else:
                all_tokens = completion_tokens

            all_token_strs = [self.detokenize([token]) for token in all_tokens]
            all_logprobs = [
                self.logits_to_logprobs(row.tolist()) for row in self.model.scores
            ][token_offset:]
            for token, token_str, logprobs_token in zip(
                all_tokens, all_token_strs, all_logprobs
            ):
                text_offsets.append(text_offset)
                text_offset += len(token_str)
                tokens.append(token_str)
                sorted_logprobs = list(
                    sorted(
                        zip(logprobs_token, range(len(logprobs_token))), reverse=True
                    )
                )
                token_logprobs.append(sorted_logprobs[int(token)][0])
                top_logprob: Optional[Dict[str, float]] = {
                    self.detokenize([i]): logprob
                    for logprob, i in sorted_logprobs[:logprobs]
                }
                top_logprob.update({token_str: logprobs_token[int(token)]})
                top_logprobs.append(top_logprob)
            # Weird idosincracy of the OpenAI API where
            # token_logprobs and top_logprobs are null for
            # the first token.
            if echo and len(all_tokens) > 0:
                token_logprobs[0] = None
                top_logprobs[0] = None
            logprobs_or_none = {
                "tokens": tokens,
                "text_offset": text_offsets,
                "token_logprobs": token_logprobs,
                "top_logprobs": top_logprobs,
            }

        yield {
            "id": completion_id,
            "object": "text_completion",
            "created": created,
            "model": model_name,
            "choices": [
                {
                    "text": text_str,
                    "index": 0,
                    "logprobs": logprobs_or_none,
                    "finish_reason": finish_reason,
                }
            ],
            "usage": {
                "prompt_tokens": len(prompt_tokens),
                "completion_tokens": len(completion_tokens),
                "total_tokens": len(prompt_tokens) + len(completion_tokens),
            },
        }

    def create_completion(
        self,
        prompt: str,
        suffix: Optional[str] = None,
        max_tokens: int = 128,
        temperature: float = 0.8,
        top_p: float = 0.95,
        logprobs: Optional[int] = None,
        echo: bool = False,
        stop: Optional[Union[str, List[str]]] = [],
        frequency_penalty: float = 0.0,
        presence_penalty: float = 0.0,
        stream: bool = False,
        model: Optional[str] = None,
    ) -> Union[Completion, Iterator[CompletionChunk]]:
        """Generate text from a prompt.

        Args:
            prompt: The prompt to generate text from.
            suffix: A suffix to append to the generated text. If None, no suffix is appended.
            max_tokens: The maximum number of tokens to generate.
            temperature: The temperature to use for sampling.
            top_p: The top-p value to use for sampling.
            logprobs: The number of logprobs to return. If None, no logprobs are returned.
            echo: Whether to echo the prompt.
            stop: A list of strings to stop generation when encountered.
            repeat_penalty: The penalty to apply to repeated tokens.
            top_k: The top-k value to use for sampling.
            stream: Whether to stream the results.

        Raises:
            ValueError: If the requested tokens exceed the context window.
            RuntimeError: If the prompt fails to tokenize or the model fails to evaluate the prompt.

        Returns:
            Response object containing the generated text.
        """
        completion_or_chunks = self._create_completion(
            prompt=prompt,
            suffix=suffix,
            max_tokens=max_tokens,
            temperature=temperature,
            top_p=top_p,
            logprobs=logprobs,
            echo=echo,
            stop=stop,
            frequency_penalty=frequency_penalty,
            presence_penalty=presence_penalty,
            stream=stream,
            model=model,
        )
        if stream:
            chunks: Iterator[CompletionChunk] = completion_or_chunks
            return chunks
        completion: Completion = next(completion_or_chunks)  # type: ignore
        return completion

    def eos_token(self):
        return self.model.eos_token()

    def logits_to_logprobs(
        self, logits: npt.NDArray[np.float32]
    ) -> npt.NDArray[np.float32]:
        return np.exp(logits) / (np.sum(np.exp(logits)))  # type: ignore


class Settings(BaseSettings):
    model_file: str
    n_gpu_layers: int = 32
    n_batch: int = 2048
    n_threads: int = max(multiprocessing.cpu_count() // 2, 1)
    sentencepiece_model: Optional[str] = None


class CreateCompletionRequest(BaseModel):
    prompt: Union[str, List[str]] = Field(
        default="", description="The prompt to generate completions for."
    )
    suffix: Optional[str] = Field(
        default=None,
        description="A suffix to append to the generated text. If None, no suffix is appended. Useful for chatbots.",
    )
    max_tokens: int = Field(
        default=16,
        ge=1,
        le=2048,
        description="The maximum number of tokens to generate.",
    )
    temperature: float = Field(
        default=0.8,
        ge=0.0,
        le=2.0,
        description="Adjust the randomness of the generated text.\n\n"
        + "Temperature is a hyperparameter that controls the randomness of the generated text. It affects the probability distribution of the model's output tokens. A higher temperature (e.g., 1.5) makes the output more random and creative, while a lower temperature (e.g., 0.5) makes the output more focused, deterministic, and conservative. The default value is 0.8, which provides a balance between randomness and determinism. At the extreme, a temperature of 0 will always pick the most likely next token, leading to identical outputs in each run.",
    )
    top_p: float = Field(
        default=0.95,
        ge=0.0,
        le=1.0,
        description="Limit the next token selection to a subset of tokens with a cumulative probability above a threshold P.\n\n"
        + "Top-p sampling, also known as nucleus sampling, is another text generation method that selects the next token from a subset of tokens that together have a cumulative probability of at least p. This method provides a balance between diversity and quality by considering both the probabilities of tokens and the number of tokens to sample from. A higher value for top_p (e.g., 0.95) will lead to more diverse text, while a lower value (e.g., 0.5) will generate more focused and conservative text.",
    )
    echo: bool = Field(
        default=False,
        description="Whether to echo the prompt in the generated text. Useful for chatbots.",
    )
    stop: Optional[Union[str, List[str]]] = Field(
        default=None,
        description="A list of tokens at which to stop generation. If None, no stop tokens are used.",
    )
    stream: bool = Field(
        default=False,
        description="Whether to stream the results as they are generated. Useful for chatbots.",
    )
    logprobs: Optional[int] = Field(
        default=None,
        ge=0,
        description="The number of logprobs to generate. If None, no logprobs are generated.",
    )
    presence_penalty: Optional[float] = Field(
        default=0.0,
        ge=-2.0,
        le=2.0,
        description="Positive values penalize new tokens based on whether they appear in the text so far, increasing the model's likelihood to talk about new topics.",
    )
    frequency_penalty: Optional[float] = Field(
        default=0.0,
        ge=-2.0,
        le=2.0,
        description="Positive values penalize new tokens based on their existing frequency in the text so far, decreasing the model's likelihood to repeat the same line verbatim.",
    )
    logprobs: Optional[int] = Field(None)

    # ignored or currently unsupported
    model: Optional[str] = Field(
        description="The model to use for generating completions."
    )
    n: Optional[int] = 1
    best_of: Optional[int] = 1
    logit_bias: Optional[Dict[str, float]] = Field(None)
    user: Optional[str] = Field(None)

    class Config:
        schema_extra = {
            "example": {
                "prompt": "def fib(n):",
                "stop": ["\n\n"],
                "temperature": 0,
                "max_tokens": 34,
            }
        }


settings = Settings(model_file=environ.get("MODEL"))  # type: ignore
app = FastAPI(
    title="Code Completion API",
    description="""
## Editor Setup

### VSCode

Add the following to your `settings.json`:

```json
{
    "github.copilot.advanced": {
        "debug.testOverrideProxyUrl": "http://localhost:8000",
        "debug.overrideProxyUrl": "http://localhost:8000"
    }
}
```

### Vim / Neovim

Add the following to your vimrc or init.vim:

```
let g:copilot_proxy = 'localhost:8000'
let g:copilot_strict_ssl = 0
```
""",
)
outer_lock = Lock()
inner_lock = Lock()

tokenizer = (
    ReplitSentencepieceTokenizer(settings.sentencepiece_model)
    if settings.sentencepiece_model
    else None
)


def cancel_callback():
    return outer_lock.locked()


model = OpenAIify(
    ReplitModel.init_from_file(
        model_file=settings.model_file,
        n_gpu_layers=settings.n_gpu_layers,
        tokenizer=tokenizer,
        cancel_callback=cancel_callback,
    ),
    # check if any other requests are pending in the same thread and cancel the stream if so
    cancel_callback=cancel_callback,
)


def get_model():
    # NOTE: This double lock allows the currently streaming model to check
    # if any other requests are pending in the same thread and cancel the
    # stream if so.
    outer_lock.acquire()
    release_outer_lock = True
    try:
        inner_lock.acquire()
        try:
            outer_lock.release()
            release_outer_lock = False
            yield model
        finally:
            inner_lock.release()
    finally:
        if release_outer_lock:
            outer_lock.release()


# Used to support copilot.vim
@app.get("/copilot_internal/v2/token")
def get_copilot_token():
    content = {"token": "1", "expires_at": 2600000000, "refresh_in": 900}
    return dict(status_code=200, content=content)


CreateCompletionResponse = create_model_from_typeddict(Completion)


# Used to support copilot.vim
@app.post(
    "/v1/engines/copilot-codex/completions",
    # response_model=CreateCompletionResponse,
)
@app.post(
    "/v1/completions",
    # response_model=CreateCompletionResponse,
)
async def create_completion(
    request: Request,
    body: CreateCompletionRequest,
    model: ReplitModel = Depends(get_model),
):
    if isinstance(body.prompt, list):
        assert len(body.prompt) <= 1
        body.prompt = body.prompt[0] if len(body.prompt) > 0 else ""

    exclude = {
        "n",
        "best_of",
        "logit_bias",
        "user",
    }
    kwargs = body.dict(exclude=exclude)
    if body.stream:
        send_chan, recv_chan = anyio.create_memory_object_stream(10)

        async def event_publisher(
            inner_send_chan: MemoryObjectSendStream[Dict[str, Union[str, bool]]]
        ):
            async with inner_send_chan:
                try:
                    iterator: Iterator[CompletionChunk] = await run_in_threadpool(model.create_completion, **kwargs)  # type: ignore
                    async_iterator: AsyncIterator[
                        CompletionChunk
                    ] = iterate_in_threadpool(iterator)
                    async for chunk in async_iterator:
                        await inner_send_chan.send(dict(data=json.dumps(chunk)))
                        if await request.is_disconnected():
                            raise anyio.get_cancelled_exc_class()()
                    await inner_send_chan.send(dict(data="[DONE]"))
                except anyio.get_cancelled_exc_class() as e:
                    print("disconnected")
                    with anyio.move_on_after(1, shield=True):
                        print(
                            f"Disconnected from client (via refresh/close) {request.client}"
                        )
                        await inner_send_chan.send(dict(closing=True))
                        raise e

        return EventSourceResponse(
            recv_chan, data_sender_callable=partial(event_publisher, send_chan)
        )
    else:
        completion: Completion = await run_in_threadpool(model.create_completion, **kwargs)  # type: ignore
        return completion
